Package: aws.signature
Type: Package
Title: Amazon Web Services Request Signatures
Version: 0.6.0
Date: 2020-06-01
Authors@R: c(person("Thomas J.", "Leeper",
                    role = c("aut"),
                    email = "thosjleeper@gmail.com",
                    comment = c(ORCID = "0000-0003-4097-6326")),
             person("Jonathan", "Stott", email = "jonathan.stott@magairports.com", role = c("cre", "aut")),
             person("Mike", "Kaminsky", email = "kaminsky.michael@gmail.com", role = "ctb")
            )
Description: Generates version 2 and version 4 request signatures for Amazon Web Services ('AWS') <https://aws.amazon.com/> Application Programming Interfaces ('APIs') and provides a mechanism for retrieving credentials from environment variables, 'AWS' credentials files, and 'EC2' instance metadata. For use on 'EC2' instances, users will need to install the suggested package 'aws.ec2metadata' <https://cran.r-project.org/package=aws.ec2metadata>.
License: GPL (>= 2)
Imports: digest, base64enc
Suggests: testthat (>= 2.1.0), aws.ec2metadata (>= 0.1.6)
URL: https://github.com/cloudyr/aws.signature
BugReports: https://github.com/cloudyr/aws.signature/issues
RoxygenNote: 7.1.0
NeedsCompilation: no
Packaged: 2020-06-01 09:51:14 UTC; jon_m
Author: Thomas J. Leeper [aut] (<https://orcid.org/0000-0003-4097-6326>),
  Jonathan Stott [cre, aut],
  Mike Kaminsky [ctb]
Maintainer: Jonathan Stott <jonathan.stott@magairports.com>
Repository: CRAN
Date/Publication: 2020-06-01 10:10:02 UTC
Built: R 4.3.3; ; 2025-01-24 04:02:10 UTC; unix

Package: aws.signature
Type: Package
Title: Amazon Web Services Request Signatures
Version: 0.6.2
Date: 2020-06-01
Authors@R: c(person("Thomas J.", "Leeper",
                    role = "aut",
                    email = "thosjleeper@gmail.com",
                    comment = c(ORCID = "0000-0003-4097-6326")),
             person("Jonathan", "Stott", email = "jonathan.stott@magairports.com", role = "aut"),
             person("Gonzalo", "Diaz", email = "gonzalomartindiaz22@gmail.com", role = c("cre", "aut"),
                    comment = c(ORCID = "0009-0003-9056-4896")),
             person("Mike", "Kaminsky", email = "kaminsky.michael@gmail.com", role = "ctb"),
             person("Mark", "Douthwaite", email = "mark.douthwaite@peak.ai", role = "ctb"),
             person("Jason", "Gofford", email = "jason.gofford@peak.ai", role = "ctb"),
             person("Luke", "Dyer", email = "luke.dyer@peak.ai", role = "ctb")
            )
Description: Generates version 2 and version 4 request signatures for Amazon Web Services ('AWS') <https://aws.amazon.com/> Application Programming Interfaces ('APIs') and provides a mechanism for retrieving credentials from environment variables, 'AWS' credentials files, and 'EC2' instance metadata. For use on 'EC2' instances, users will need to install the suggested package 'aws.ec2metadata' <https://cran.r-project.org/package=aws.ec2metadata>.
License: GPL (>= 2)
Encoding: UTF-8
Imports: digest, base64enc, jsonlite, curl
Suggests: devtools, testthat (>= 2.1.0), aws.ec2metadata (>= 0.1.6)
URL: https://github.com/cloudyr/aws.signature
BugReports: https://github.com/cloudyr/aws.signature/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-18 03:25:52 UTC; gonzalo
Author: Thomas J. Leeper [aut] (ORCID: <https://orcid.org/0000-0003-4097-6326>),
  Jonathan Stott [aut],
  Gonzalo Diaz [cre, aut] (ORCID:
    <https://orcid.org/0009-0003-9056-4896>),
  Mike Kaminsky [ctb],
  Mark Douthwaite [ctb],
  Jason Gofford [ctb],
  Luke Dyer [ctb]
Maintainer: Gonzalo Diaz <gonzalomartindiaz22@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-21 05:40:02 UTC
Built: R 4.5.2; ; 2026-07-21 05:50:41 UTC; unix

# Generated by roxygen2: do not edit by hand

export(assume_role_with_web_identity)
export(canonical_request)
export(default_credentials_file)
export(locate_credentials)
export(read_credentials)
export(signature_v2_auth)
export(signature_v4)
export(signature_v4_auth)
export(string_to_sign)
export(use_credentials)
importFrom(base64enc,base64encode)
importFrom(digest,digest)
importFrom(digest,hmac)

# Amazon Web Services Request Signatures

**aws.signature** is a package for creating request signatures for Amazon Web Services (AWS) APIs. It supports both the current [Signature Version 4](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html) and the legacy [Signature Version 2](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html). The former is used by most services. The high-level functions `signature_v4_auth()` and `signature_v2_auth()` translate request parameters into appropriate HTTP Authorization headers to pass to the APIs.

To use the package, you will need an AWS account and to enter your credentials into R. Your keypair can be generated on the [IAM Management Console](https://aws.amazon.com/) under the heading *Access Keys*. Note that you only have access to your secret key once. After it is generated, you need to save it in a secure location. New keypairs can be generated at any time if yours has been lost, stolen, or forgotten. The [**aws.iam** package](https://github.com/cloudyr/aws.iam) profiles tools for working with IAM, including creating roles, users, groups, and credentials programmatically; it is not needed to *use* IAM credentials.

By default, when loaded the package checks for environment variables. If absent, it checks for a default credentials file and loads credentials from it into environment variables; the profile used from that file can be regulated by setting the `AWS_PROFILE` environment variable before loading this package (the `"default" profile is assumed if none is specified). This means the package and any dependencies should *just work* without needing to explicitly set or pass credentials within R code.

Regardless of this initial configuration, all **awspack** packages allow the use of credentials specified in a number of ways, in the following priority order:

 1. User-supplied values passed directly to functions.
 2. Environment variables, which can alternatively be set on the command line prior to starting R or via an `Renviron.site` or `.Renviron` file, which are used to set environment variables in R during startup (see `? Startup`). Or they can be set within R:
 
    ```R
    Sys.setenv("AWS_ACCESS_KEY_ID" = "mykey",
               "AWS_SECRET_ACCESS_KEY" = "mysecretkey",
               "AWS_DEFAULT_REGION" = "us-east-1",
               "AWS_SESSION_TOKEN" = "mytoken")
    ```
 3. If R is running on an EC2 instance, the role profile credentials provided by [**aws.ec2metadata**](https://cran.r-project.org/package=aws.ec2metadata), *if the **aws.ec2metadata** package is installed*.
 4. If R is running on an ECS task, the role profile credentials provided by [**aws.ec2metadata**](https://cran.r-project.org/package=aws.ec2metadata), *if the **aws.ec2metadata** package is installed*.
 5. Profiles saved in a `/.aws/credentials` "dot file" in the current working directory. The profile used can be regulated by the `AWS_PROFILE` environment variable, otherwise the `"default" profile is assumed if none is specified or the specified profile is missing.
 6. [A centralized credentials file](https://blogs.aws.amazon.com/security/post/Tx3D6U6WSFGOK2H/A-New-and-Standardized-Way-to-Manage-Credentials-in-the-AWS-SDKs), containing credentials for multiple accounts. The location of this file is given by the `AWS_SHARED_CREDENTIALS_FILE` environment variable or, if that is missing, by `~/.aws/credentials` (or an OS-specific equivalent). The profile used from that file can be regulated by the `AWS_PROFILE` environment variable, otherwise the `"default" profile is assumed if none is specified or the specified profile is missing.

Because all functions requesting a signature walk this entire list of potential credentials sources, it typically makes sense to set environment variables otherwise a potentially large performance penalty can be paid. For this reason, it is usually better to explicitly invoke a profiles stored in a local or centralized (e.g., `~/.aws/credentials`) credentials file using:

```R
# use your 'default' account credentials
use_credentials()

# use an alternative credentials profile
use_credentials(profile = "bob")
```

For purposes of debugging, it can be useful to set the `verbose = TRUE` argument (or globally set `options(verbose = TRUE)`) in order to see what values are being used for signing requests.

Temporary session tokens are stored in environment variable `AWS_SESSION_TOKEN` (and will be stored there by the `use_credentials()` function). The [aws.iam package](https://github.com/cloudyr/aws.iam/) provides an R interface to IAM roles and the generation of temporary session tokens via the security token service (STS). On EC2 instances or ECS tasks, the [**aws.ec2metadata**](https://cran.r-project.org/package=aws.ec2metadata) package should be installed so that signatures are signed with appropriate, dynamically updated credentials.

As a fail safe the `us-east-1` region is used whenever a region is not found.

## Installation

[![CRAN](https://www.r-pkg.org/badges/version/aws.signature)](https://cran.r-project.org/package=aws.signature)
![Downloads](https://cranlogs.r-pkg.org/badges/aws.signature)
[![Build Status](https://travis-ci.org/cloudyr/aws.signature.png?branch=master)](https://app.travis-ci.com/cloudyr/aws.signature) 
[![codecov.io](https://codecov.io/github/cloudyr/aws.signature/coverage.svg?branch=master)](https://app.codecov.io/github/cloudyr/aws.signature?branch=master)

To install the latest package version, it is recommended to install from the cloudyr drat repository:

```R
# latest stable version
install.packages("aws.signature", repos = c(cloudyr = "http://cloudyr.github.io/drat", getOption("repos")))
```

Or, to pull a potentially unstable version directly from GitHub:

```R
if (!require("remotes")) {
    install.packages("remotes")
}
remotes::install_github("cloudyr/aws.signature")
```

To install the latest version from CRAN, simply use `install.packages("aws.signature")`.

---
[![cloudyr project logo](http://i.imgur.com/JHS98Y7.png)](https://github.com/cloudyr)

.onLoad <- function(libname, pkgname) {
    # If credentials are in environment variables, use those.
    creds <- locate_credentials()
    if (!all(is.null(creds$key), is.null(creds$secret))) {
        return(invisible(NULL))
    }
    # Load default AWS credentials allowing the package to behave like the AWS CLI
    if (file.exists(default_credentials_file())) {
        use_credentials()
    }
    return(invisible(NULL))
}

#' @title Construct a Canonical Request
#' @description Construct a Canonical Request from request elements
#' @param verb A character string containing the HTTP verb being used in the request.
#' @param canonical_uri A character string containing the \dQuote{canonical URI}, meaning the contents of the API request URI excluding the host and the query parameters.
#' @param query_args A named list of character strings containing the query string values (if any) used in the API request.
#' @param canonical_headers A named list of character strings containing the headers used in the request.
#' @param request_body The body of the HTTP request, or a filename. If a filename, hashing is performed on the file without reading it into memory.
#' @param signed_body Sign the body request and add the correct header (x-amz-content-sha256) to the list of headers
#' @details This function creates a \dQuote{Canonical Request}, which is part of the Signature Version 4. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
#' @return A list containing
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references
#'   \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Create a Canonical Request For Signature Version 4}
#' @seealso \code{\link{signature_v4}}, \code{\link{signature_v4_auth}}
#' @examples
#' # From AWS documentation
#' # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html
#' fromDocs <- "POST
#' /
#'
#' content-type:application/x-www-form-urlencoded; charset=utf-8
#' host:iam.amazonaws.com
#' x-amz-date:20110909T233600Z
#'
#' content-type;host;x-amz-date
#' b6359072c78d70ebee1e81adcbab4f01bf2c23245fa365ef83fe8f1f955085e2"
#'
#' hdrs <- list(`Content-Type` = "application/x-www-form-urlencoded; charset=utf-8",
#'              Host = "iam.amazonaws.com",
#'              `x-amz-date` = "20110909T233600Z")
#' r <- canonical_request(verb = "POST",
#'                        canonical_uri = "/",
#'                        query_args = list(),
#'                        canonical_headers = hdrs,
#'                        request_body = "Action=ListUsers&Version=2010-05-08")
#'
#' identical(fromDocs, r$canonical)
#' @seealso \code{link{signature_v4_aut}}, \code{\link{string_to_sign}}
#' @export
canonical_request <-
    function(
        verb,
        canonical_uri = "",
        query_args = list(),
        canonical_headers,
        request_body = "",
        signed_body = FALSE
    ) {
        if (is.character(request_body) && file.exists(request_body)) {
            body_hash <- tolower(digest::digest(
                request_body,
                file = TRUE,
                algo = "sha256",
                serialize = FALSE
            ))
        } else {
            body_hash <- tolower(digest::digest(
                request_body,
                algo = "sha256",
                serialize = FALSE
            ))
        }

        # set sort locale
        lc <- Sys.getlocale(category = "LC_COLLATE")
        Sys.setlocale(category = "LC_COLLATE", locale = "C")
        on.exit(Sys.setlocale(category = "LC_COLLATE", locale = lc))

        if (isTRUE(signed_body)) {
            canonical_headers[['x-amz-content-sha256']] <- body_hash
        }

        names(canonical_headers) <- tolower(names(canonical_headers))
        canonical_headers <- canonical_headers[order(names(canonical_headers))]
        # trim leading, trailing, and all non-quoted duplicated spaces
        trimmed_headers <- gsub(
            "[[:space:]]{2,}",
            " ",
            trimws(canonical_headers)
        )
        header_string <- paste0(
            names(canonical_headers),
            ":",
            trimmed_headers,
            "\n",
            collapse = ""
        )
        signed_headers <- paste(
            names(canonical_headers),
            sep = "",
            collapse = ";"
        )
        if (length(query_args)) {
            query_args <- unlist(query_args[order(names(query_args))])
            a <- paste0(
                sapply(names(query_args), URLencode, reserved = TRUE),
                "=",
                sapply(as.character(query_args), URLencode, reserved = TRUE)
            )
            query_string <- paste(a, sep = "", collapse = "&")
        } else {
            query_string <- ""
        }
        out <- paste(
            verb,
            canonical_uri,
            query_string,
            header_string,
            signed_headers,
            body_hash,
            sep = "\n"
        )
        return(list(
            headers = signed_headers,
            body = body_hash,
            canonical = out,
            hash = digest::digest(out, algo = "sha256", serialize = FALSE)
        ))
    }

context("AWS Example Test Suite")
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-test-suite.html

test_that("AWS test suite via canonical_request", {
    ex <- "GET
/
foo=Zoo&foo=aha
date:Mon, 09 Sep 2011 23:36:00 GMT
host:host.foo.com

date;host
e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"

    r <- canonical_request(verb = "GET",
                           canonical_uri = "/",
                           query_args = list(foo = "Zoo", foo = "aha"),
                           canonical_headers = list(host = "host.foo.com",
                                                    date = "Mon, 09 Sep 2011 23:36:00 GMT"),
                           request_body = "")
    expect_identical(r$canonical, ex, label = "Canonical request matches")
})

test_that("AWS test suite via string_to_sign", {
    ex <- "AWS4-HMAC-SHA256
20110909T233600Z
20110909/us-east-1/host/aws4_request
e25f777ba161a0f1baf778a87faf057187cf5987f17953320e3ca399feb5f00d"
    s <- string_to_sign(algorithm = "AWS4-HMAC-SHA256",
         datetime = "20110909T233600Z",
         region = "us-east-1",
         service = "host",
         request_hash = "e25f777ba161a0f1baf778a87faf057187cf5987f17953320e3ca399feb5f00d")
    expect_identical(s, ex, label = "String to sign matches")
})

test_that("AWS test suite via signature_v4", {
    tosign <- "AWS4-HMAC-SHA256
20110909T233600Z
20110909/us-east-1/host/aws4_request
e25f777ba161a0f1baf778a87faf057187cf5987f17953320e3ca399feb5f00d"
    ex <- "be7148d34ebccdc6423b19085378aa0bee970bdc61d144bd1a8c48c33079ab09"
    s <- signature_v4(secret = "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY",
                      date = "20110909",
                      region = "us-east-1",
                      service = "host",
                      string_to_sign = tosign, verbose = TRUE)
    expect_identical(s, ex, label = "String to sign matches")
})

test_that("AWS test suite via signature_v4_auth", {
    d <- "20110909T233600Z"
    s <- signature_v4_auth(datetime = d,
                           region = "us-east-1",
                           service = "host",
                           verb = "GET",
                           action = "/",
                           query_args = list(foo = "Zoo", foo = "aha"),
                           canonical_headers = list(host = "host.foo.com",
                                                    date = d),
                           request_body = "",
                           key = "AKIDEXAMPLE",
                           secret = "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY",
                           query = FALSE,
                           algorithm = "AWS4-HMAC-SHA256")
    expect_identical(s$Credential, "AKIDEXAMPLE/20110909/us-east-1/host/aws4_request", 
                     label = "Credential string matches")
    expect_identical(s$BodyHash, "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855", 
                     label = "Body hash string matches")
    expect_identical(s$SignedHeaders, "date;host", 
                     label = "Signed header string matches")
    #expect_identical(s$Signature, "be7148d34ebccdc6423b19085378aa0bee970bdc61d144bd1a8c48c33079ab09", 
    #                 label = "Signature string matches")
    #expect_identical(s$SignatureHeader, "AWS4-HMAC-SHA256 Credential=AKIDEXAMPLE/20110909/us-east-1/host/aws4_request, SignedHeaders=date;host, Signature=be7148d34ebccdc6423b19085378aa0bee970bdc61d144bd1a8c48c33079ab09", 
    #                 label = "Full authorization string matches")
    
})

test_that("AWS test suite via signature_v2_auth", {
true_string <- "GET
elasticmapreduce.amazonaws.com
/
AWSAccessKeyId=AKIAIOSFODNN7EXAMPLE&Action=DescribeJobFlows&SignatureMethod=HmacSHA256&SignatureVersion=2&Timestamp=2011-10-03T15%3A19%3A30&Version=2009-03-31"
    true_sig <- "i91nKc4PWAt0JJIdXwz9HxZCJDdiy6cf/Mj6vPxyYIs="

    q1 <- 
    list(Action = "DescribeJobFlows",
         Version = "2009-03-31",
         AWSAccessKeyId = "AKIAIOSFODNN7EXAMPLE",
         SignatureVersion = "2",
         SignatureMethod = "HmacSHA256",
         Timestamp = "2011-10-03T15:19:30")

    sig1 <- 
    signature_v2_auth(datetime = "2011-10-03T15:19:30",
                      service = "elasticmapreduce.amazonaws.com",
                      verb = "GET",
                      path = "/",
                      query_args = q1,
                      key = q1$AWSAccessKeyId,
                      secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
    expect_identical(true_string, sig1$CanonicalRequest, label = "v2 CanonicalRequest correct")
    expect_identical(true_sig, sig1$Signature, label = "v2 Signature correct")

    q2 <- 
    list(Action = "DescribeJobFlows",
         Version = "2009-03-31",
         Timestamp = "2011-10-03T15:19:30")
    sig2 <- 
    signature_v2_auth(datetime = "2011-10-03T15:19:30",
                      service = "elasticmapreduce.amazonaws.com",
                      verb = "GET",
                      path = "/",
                      query_args = q2,
                      key = "AKIAIOSFODNN7EXAMPLE",
                      secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
    expect_identical(true_string, sig2$CanonicalRequest, label = "v2 CanonicalRequest correct, setting defaults")
    expect_identical(true_sig, sig2$Signature, label = "v2 Signature correct, setting defaults")
})

test_that("session_token returned in signature_v4()", {
    s <- signature_v4_auth(datetime = "20110909T233600Z",
                           region = "us-east-1",
                           service = "host",
                           verb = "GET",
                           action = "/",
                           query_args = list(foo = "Zoo", foo = "aha"),
                           canonical_headers = list(host = "host.foo.com",
                                                    date = "20110909T233600Z"),
                           request_body = "",
                           key = "AKIDEXAMPLE",
                           secret = "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY",
                           session_token = "foobar",
                           query = FALSE,
                           algorithm = "AWS4-HMAC-SHA256")
    expect_match(s$SignedHeaders, "x-amz-security-token")
})


context("Miscellaneous Functionality")

test_that("get_ec2_role() works", {
    skip_on_cran()
    if (!requireNamespace("aws.ec2metadata", quietly = TRUE)) {
        expect_null(aws.signature:::get_ec2_role(), label = "get_ec2_role() returns NULL")
    }
})

test_that("A canonical request is generated", {
  fromDocs <- "POST
/

content-type:application/x-www-form-urlencoded; charset=utf-8
host:iam.amazonaws.com
x-amz-date:20110909T233600Z

content-type;host;x-amz-date
b6359072c78d70ebee1e81adcbab4f01bf2c23245fa365ef83fe8f1f955085e2"

  hdrs <- list(`Content-Type` = "application/x-www-form-urlencoded; charset=utf-8",
               Host = "iam.amazonaws.com",
               `x-amz-date` = "20110909T233600Z")
  r <- canonical_request(verb = "POST",
                         canonical_uri = "/",
                         query_args = list(),
                         canonical_headers = hdrs,
                         request_body = "Action=ListUsers&Version=2010-05-08")

  expect_identical(r$canonical, fromDocs)
})


test_that("A canonical request is generated with a signed body", {
  fromDocs <- "POST
/

content-type:application/x-www-form-urlencoded; charset=utf-8
host:iam.amazonaws.com
x-amz-content-sha256:b6359072c78d70ebee1e81adcbab4f01bf2c23245fa365ef83fe8f1f955085e2
x-amz-date:20110909T233600Z

content-type;host;x-amz-content-sha256;x-amz-date
b6359072c78d70ebee1e81adcbab4f01bf2c23245fa365ef83fe8f1f955085e2"
  
  hdrs <- list(`Content-Type` = "application/x-www-form-urlencoded; charset=utf-8",
               Host = "iam.amazonaws.com",
               `x-amz-date` = "20110909T233600Z")
  r <- canonical_request(verb = "POST",
                         canonical_uri = "/",
                         query_args = list(),
                         canonical_headers = hdrs,
                         signed_body = TRUE,
                         request_body = "Action=ListUsers&Version=2010-05-08")
  
  expect_identical(r$canonical, fromDocs)
})

context("Digital Ocean Example Test Suite")
# https://developers.digitalocean.com/documentation/spaces/#authentication

test_that("Digital Ocean test suite via canonical_request", {
    ex <- "GET
/
acl=
host:static-images.nyc3.digitaloceanspaces.com
x-amz-content-sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
x-amz-date:20170804T221549Z

host;x-amz-content-sha256;x-amz-date
e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"

    r <- canonical_request(verb = "GET",
                           canonical_uri = "/",
                           query_args = list(acl = ""),
                           canonical_headers = list(host = "static-images.nyc3.digitaloceanspaces.com",
                                                    'x-amz-content-sha256' = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                                                    'x-amz-date' = "20170804T221549Z"),
                           request_body = "")
    expect_identical(r$canonical, ex, label = "Digital Ocean canonical request matches")
})

#' @title Signature Version 2
#' @description Generates AWS Signature Version 2
#' @param datetime A character string containing a date in the form of \dQuote{YYYY-MM-DDTH:M:S}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.
#' @param verb A character string specify an HTTP verb/method (e.g., \dQuote{GET}).
#' @param service A character string containing the full hostname of an AWS service (e.g., \dQuote{iam.amazonaws.com}, etc.)
#' @param path A character string specify the path to the API endpoint.
#' @param query_args A list containing named query arguments.
#' @template key
#' @template secret
#' @template region
#' @template force_credentials
#' @template verbose
#' @details This function generates an AWS Signature Version 2 for authorizing API requests. The function returns both an updated set of query string parameters, containing the required signature-related entries, as well as a \code{Signature} field containing the Signature string itself. Version 2 is mostly deprecated and in most cases users should rely on \code{\link{signature_v4_auth}} for Version 4 signatures instead.
#' @return A list.
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html}{AWS General Reference: Signature Version 2 Signing Process}
#' @examples
#' \dontrun{
#' # examples from:
#' # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html
#'
#' true_string <- paste0("GET\n",
#' "elasticmapreduce.amazonaws.com\n",
#' "/\n",
#' "AWSAccessKeyId=AKIAIOSFODNN7EXAMPLE",
#' "&Action=DescribeJobFlows",
#' "&SignatureMethod=HmacSHA256",
#' "&SignatureVersion=2",
#' "&Timestamp=2011-10-03T15\%3A19\%3A30",
#' "&Version=2009-03-31", collapse = "")
#' true_sig <- "i91nKc4PWAt0JJIdXwz9HxZCJDdiy6cf/Mj6vPxyYIs="
#'
#' q1 <-
#' list(Action = "DescribeJobFlows",
#'      Version = "2009-03-31",
#'      AWSAccessKeyId = "AKIAIOSFODNN7EXAMPLE",
#'      SignatureVersion = "2",
#'      SignatureMethod = "HmacSHA256",
#'      Timestamp = "2011-10-03T15:19:30")
#'
#' sig1 <-
#' signature_v2_auth(datetime = "2011-10-03T15:19:30",
#'                   service = "elasticmapreduce.amazonaws.com",
#'                   verb = "GET",
#'                   path = "/",
#'                   query_args = q1,
#'                   key = q1$AWSAccessKeyId,
#'                   secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
#' identical(true_string, sig1$CanonicalRequest)
#' identical(true_sig, sig1$Signature)
#'
#' # leaving out some defaults
#' q2 <-
#' list(Action = "DescribeJobFlows",
#'      Version = "2009-03-31",
#'      Timestamp = "2011-10-03T15:19:30")
#' sig2 <-
#' signature_v2_auth(datetime = "2011-10-03T15:19:30",
#'                   service = "elasticmapreduce.amazonaws.com",
#'                   verb = "GET",
#'                   path = "/",
#'                   query_args = q2,
#'                   key = "AKIAIOSFODNN7EXAMPLE",
#'                   secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
#' identical(true_string, sig2$CanonicalRequest)
#' identical(true_sig, sig2$Signature)
#' }
#' @seealso \code{\link{signature_v4_auth}}, \code{\link{use_credentials}}
#' @importFrom digest digest hmac
#' @importFrom base64enc base64encode
#' @export
signature_v2_auth <-
    function(
        datetime = format(Sys.time(), "%Y-%m-%dT%H:%M:%S", tz = "UTC"),
        verb,
        service,
        path,
        query_args = list(),
        key = NULL,
        secret = NULL,
        region = NULL,
        force_credentials = FALSE,
        verbose = getOption("verbose", FALSE)
    ) {
        if (isTRUE(force_credentials)) {
            credentials <- list(key = key, secret = secret, region = region)
            if (isTRUE(verbose)) {
                if (!is.null(key)) {
                    message("Using user-supplied value for AWS Access Key ID")
                }
                if (!is.null(secret)) {
                    message(
                        "Using user-supplied value for AWS Secret Access Key"
                    )
                }
                if (!is.null(region)) {
                    message(sprintf(
                        "Using user-supplied value for AWS Region ('%s')",
                        region
                    ))
                }
            }
        } else {
            credentials <- locate_credentials(
                key = key,
                secret = secret,
                region = region,
                verbose = verbose
            )
        }

        # set sort locale
        lc <- Sys.getlocale(category = "LC_COLLATE")
        Sys.setlocale(category = "LC_COLLATE", locale = "C")
        on.exit(Sys.setlocale(category = "LC_COLLATE", locale = lc))

        # sort query arguments
        if (!"Timestamp" %in% names(query_args)) {
            query_args$Timestamp <- datetime
        } else {
            datetime <- query_args$Timestamp
        }
        if (!"SignatureVersion" %in% names(query_args)) {
            query_args$SignatureVersion <- "2"
        }
        if (!"SignatureMethod" %in% names(query_args)) {
            query_args$SignatureMethod <- "HmacSHA256"
        }
        if (!"AWSAccessKeyId" %in% names(query_args)) {
            query_args$AWSAccessKeyId <- credentials$key
        }
        query_to_parse <- unlist(query_args[order(names(query_args))])
        a <- paste0(
            sapply(names(query_to_parse), URLencode, reserved = TRUE),
            "=",
            sapply(as.character(query_to_parse), URLencode, reserved = TRUE)
        )
        query_string <- paste(a, sep = "", collapse = "&")

        canonical_request <- paste(
            verb,
            service,
            path,
            query_string,
            sep = "\n"
        )
        signature <- digest::hmac(
            key = credentials$secret,
            object = canonical_request,
            algo = "sha256",
            serialize = FALSE,
            raw = TRUE
        )
        sig_encoded <- base64enc::base64encode(signature)
        query_args$Signature <- sig_encoded

        # return list
        structure(
            list(
                CanonicalRequest = canonical_request,
                StringToSign = canonical_request,
                Verb = verb,
                Query = query_args,
                Service = service,
                Path = path,
                Signature = sig_encoded,
                AccessKeyId = credentials[["key"]],
                SecretAccessKey = credentials[["secret"]],
                SessionToken = credentials[["session_token"]],
                Region = credentials[["region"]]
            ),
            class = "aws_signature_v2"
        )
    }

URLencode <- function(URL, reserved = FALSE) { 
    OK <- paste0("[^", if (!reserved) 
        "][!$&'()*+,;=:/?@#", "ABCDEFGHIJKLMNOPQRSTUVWXYZ", 
        "abcdefghijklmnopqrstuvwxyz0123456789._~-", "]") 
    x <- strsplit(URL, "")[[1L]] 
    z <- grep(OK, x) 
    if (length(z)) { 
        y <- sapply(x[z], function(x) paste0("%", toupper(as.character(charToRaw(x))), collapse = "")) 
        x[z] <- y 
    } 
    paste(x, collapse = "") 
} 

#' Amazon Web Services Request Signatures
#'
#' Generates Amazon Web Services (AWS) request signatures for RESTful APIs.
#'
#' This package contains functions mostly intended for developers to use in
#' building API client packages for Amazon Web Services APIs.
#'
#' The main function of interest is [signature_v4_auth()], which wraps the
#' other internal functions and returns a named list of elements to be used in
#' authenticating an API request using AWS Signature Version 4. Another
#' function, [signature_v2_auth()], implements the older, mostly deprecated
#' Version 2 algorithm.
#'
#' Recent versions of the package (>= 0.2.8) identify credentials by walking
#' through a tree of possible sources of values (described in
#' [locate_credentials()]), with optional verbosity, in a manner similar to the
#' Python boto3 library.
#'
#' A lower-level function that may be of use to end users is
#' [use_credentials()], which sets the environment variables used by this
#' package based upon values specified in a `.aws/credentials` file. That
#' function is called by default during package load, if no environment
#' variables are set.
#'
#' To use this (and any cloudyr package) on AWS EC2 instances or ECS tasks,
#' users will also need to install the **aws.ec2metadata** package, which
#' allows [locate_credentials()] to know it is running in an instance and check
#' for relevant values.
#'
#' @keywords internal
"_PACKAGE"

#' @rdname read_credentials
#' @title Use Credentials from .aws/credentials File
#' @description Use a profile from a \samp{.aws/credentials} file
#' @param profile A character string specifying which profile to use from the file. By default, the \dQuote{default} profile is used.
#' @param file A character string containing a path to a \samp{.aws/credentials} file. By default, the standard/centralized file given by \env{AWS_SHARED_CREDENTIALS_FILE} is used, otherwise an assumed default location is assumed. For \code{use_credentials}, this can also be an object of class \dQuote{aws_credentials} (as returned by \code{use_credentials}).
#' @details \code{read_credentials} reads and parses a \samp{.aws/credentials} file into an object of class \dQuote{aws_credentials}.
#'
#' \code{use_credentials} uses credentials from a profile stored in a credentials file to set the environment variables used by this package. It is called by default during package load if the \env{AWS_ACCESS_KEY_ID} and \env{AWS_SECRET_ACCESS_KEY} environment variables are not set.
#'
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references
#'   \href{https://blogs.aws.amazon.com/security/post/Tx3D6U6WSFGOK2H/A-New-and-Standardized-Way-to-Manage-Credentials-in-the-AWS-SDKs}{Amazon blog post describing the format}
#' @seealso \code{\link{signature_v2_auth}}, \code{\link{locate_credentials}}
#' @examples
#' \dontrun{
#' # read and parse a credentials file
#' read_credentials()
#'
#' # set environment variables from a profile
#' use_credentials()
#' }
#' @export
read_credentials <-
function(
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file())
) {
    file <- path.expand(file)
    if (!file.exists(file)) {
        stop(paste0("File ", shQuote(file), " does not exist."))
    }
    char <- rawToChar(readBin(file, "raw", n = 1e5L))
    parse_credentials(char)
}

#' @rdname read_credentials
#' @export
use_credentials <-
function(
  profile = Sys.getenv("AWS_PROFILE", "default"),
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file())
) {
    if (inherits(file, "aws_credentials")) {
        x <- file
    } else {
        x <- read_credentials(file)
    }
    if ("AWS_ACCESS_KEY_ID" %in% names(x[[profile]])) {
        Sys.setenv("AWS_ACCESS_KEY_ID" = x[[profile]][["AWS_ACCESS_KEY_ID"]])
    }
    if ("AWS_SECRET_ACCESS_KEY" %in% names(x[[profile]])) {
        Sys.setenv("AWS_SECRET_ACCESS_KEY" = x[[profile]][["AWS_SECRET_ACCESS_KEY"]])
    }
    if ("AWS_SESSION_TOKEN" %in% names(x[[profile]])) {
        Sys.setenv("AWS_SESSION_TOKEN" = x[[profile]][["AWS_SESSION_TOKEN"]])
    }
    if ("AWS_DEFAULT_REGION" %in% names(x[[profile]])) {
        Sys.setenv("AWS_DEFAULT_REGION" = x[[profile]][["AWS_DEFAULT_REGION"]])
    }
    invisible(x)
}

#' @rdname read_credentials
#' @export
default_credentials_file <-
function() {
    if (.Platform[["OS.type"]] == "windows") {
        home <- Sys.getenv("USERPROFILE")
    } else {
        home <- "~"
    }
    suppressWarnings(normalizePath(file.path(home, '.aws', 'credentials')))
}

parse_credentials <- function(char) {
    char <- gsub("(?m)#.*?$", "", char, perl=TRUE)
  
    s <- c(gregexpr("\\[", char)[[1]], nchar(char)+1)
    # This regex finds the header lines which look like: [name]<any old junk>EOL
    sections <- gregexpr("(?m)^(?<header_line>\\[(?<header_contents>[^]]+)\\].*)$", char, perl=TRUE)[[1]]
    
    # the section starts from the start of the match, and the start of the line
    section_starts <- sections + attr(sections, "capture.length")[,"header_line"]
    # the section ends at the start of the next section, or the end
    section_ends   <- c(sections-1, nchar(char)+1)[-1]
    
    # header name bounds
    header_starts <- attr(sections, "capture.start")[,"header_contents"]
    header_ends   <- header_starts + attr(sections, "capture.length")[,"header_contents"] - 1
    
    # extract out our headers and sections
    headers <- substring(char, header_starts, header_ends)
    section_bodies <- substring(char, section_starts, section_ends)
    names(section_bodies) <- headers
    
    # now with the section bodies, parse our variables
    # variables MUST start at the start of a line
    variable_matches <- gregexpr("(?m)^(?<name>\\w+)[ \t]*=[ \t]*", section_bodies, perl=TRUE)
    
    creds <- mapply(process_variable, section_bodies, variable_matches, SIMPLIFY = FALSE)

    structure(creds, class = "aws_credentials")
}

process_variable <- function(text, var_match) {
  # parse out names based on the start/end of the capture
  name_starts <- attr(var_match, "capture.start")[,"name"]
  name_ends   <- name_starts + attr(var_match, "capture.length")[,"name"]
  var_names <- substring(text, name_starts, name_ends  - 1)
  
  # var values start at the end of the match (i.e after = and whitespace)
  value_starts <- var_match + attr(var_match, "match.length")
  # values end just before the next name starts
  value_ends   <- c(name_starts[-1]-1, nchar(text) + 1)
  
  var_values <- substring(text, value_starts, value_ends)
  # trim all whitespace around values.
  # This makes parsing nested sections impossible, but we're parsing and ignoring them for now
  var_values <- trimws(var_values)
  
  var_values <- as.list(var_values)
  names(var_values) <- toupper(var_names)
  
  var_values
}

context("Tests reading of a credentials file")
# http://docs.aws.amazon.com/cli/latest/reference/configure/
# https://blogs.aws.amazon.com/security/post/Tx3D6U6WSFGOK2H/A-New-and-Standardized-Way-to-Manage-Credentials-in-the-AWS-SDKs

test_that("default_credentials_file() works", {
    skip_on_cran()
    expect_is(default_credentials_file(), "character", label = "default_credentials_file() returns path")
})

test_that("read_credentials() works", {
    skip_on_cran()
    expect_is(read_credentials(file = "credentials"), "aws_credentials", label = "Read credentials correctly")
    alice <- read_credentials(file = "credentials")[["Alice"]]
    bob <- read_credentials(file = "credentials")[["Bob"]]
    expect_equal(bob[["AWS_SECRET_ACCESS_KEY"]], "Bob_secret_access_key", label = "final line value read correctly")
    expect_false(identical(alice, bob), label = "Read distinct profiles correctly")
    expect_false("[" %in% names(alice))
    expect_false("[" %in% names(bob))
})

test_that("read_credentials() works even absent EOL character", {
    skip_on_cran()
    cred <- read_credentials(file = "credentials-no-eol-char")
    expect_is(cred, "aws_credentials", label = "Read credentials successfully")
    expect_equal(cred[["Bob"]][["AWS_SECRET_ACCESS_KEY"]], "Bob_secret_access_key", label = "final line value read correctly")
})

test_that("read_credentials() fails when file is missing", {
    skip_on_cran()
    expect_error(read_credentials(file = "foo"), label = "Read credentials fails on missing file")
})


context("Tests use_credentials()")

test_that("use_credentials() sets environment variables correctly", {
    
    skip_on_cran()
    
    # save environment variables
    e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
    
    # unset environment variables
    Sys.unsetenv("AWS_ACCESS_KEY_ID")
    Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
    Sys.unsetenv("AWS_SESSION_TOKEN")
    Sys.unsetenv("AWS_DEFAULT_REGION")
    
    # tests
    use_credentials(profile = "default", file = "credentials")
    expect_equal(Sys.getenv("AWS_ACCESS_KEY_ID"), "ACCESS_KEY")
    expect_equal(Sys.getenv("AWS_SECRET_ACCESS_KEY"), "SECRET_KEY")
    expect_equal(Sys.getenv("AWS_SESSION_TOKEN"), "TOKEN")
    expect_equal(Sys.getenv("AWS_DEFAULT_REGION"), "")
    
    # restore environment variables
    do.call("Sys.setenv", as.list(e))
})

context("parse_credentials()")
test_that("parse_credentials() handles some pathological cases", {
    expect_equal(
        parse_credentials("[has ]spaces]\naws=foo\nthing=bar"),
        structure(list("has " = list(AWS = "foo", THING = "bar")), class = 'aws_credentials')
        )
    
    expect_equal(
        parse_credentials("[has spaces]\naws=foo\nthing=bar"),
        structure(list("has spaces" = list(AWS = "foo", THING = "bar")), class = 'aws_credentials')
    )
    
    expect_equal(
        parse_credentials("[spaces]\naws=foo\nthing=bar\n# baz = quux"),
        structure(list("spaces" = list(AWS = "foo", THING = "bar")), class = 'aws_credentials')
    )
    
    expect_equal(
        parse_credentials("[spaces]\naws=foo\nthing=bar\n# baz = quux\ndo = dad"),
        structure(list("spaces" = list(AWS = "foo", THING = "bar", DO = "dad")), class = 'aws_credentials')
    )
    
    
    expect_equal(
        parse_credentials("[spaces]\naws=foo\ns3 =\n   thing=bar\n[nspaces]\naws= foo\nthing   =   bar\n\n"),
        structure(list("spaces" = list(AWS = "foo", S3 = "thing=bar"),
                       "nspaces" = list(AWS="foo", THING="bar")), class = 'aws_credentials')
    )
})

#' @title Construct a String To Sign
#' @description Construct a String to Sign from request elements
#' @param algorithm A character string containing the hashing algorithm used in signing process. Should only be \dQuote{AWS4-HMAC-SHA256}.
#' @param datetime A character string containing a UTC date in the form of \dQuote{YYYYMMDDTHHMMSSZ}.
#' @param region A character string containing the AWS region for the request.
#' @param service A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).
#' @param request_hash A character string containing the hash of the canonical request, perhaps as returned by \code{\link{canonical_request}}.
#' @details This is a mostly internal function that creates a \dQuote{String To Sign}, which is part of the Signature Version 4. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references
#'   \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Create a String to Sign for Signature Version 4}
#' @examples
#' # From AWS documentation
#' rh <- "3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2"
#' sts <-
#' string_to_sign(datetime = "20110909T233600Z",
#'                region = "us-east-1",
#'                service = "iam",
#'                request_hash = rh)
#' identical(sts, "AWS4-HMAC-SHA256
#' 20110909T233600Z
#' 20110909/us-east-1/iam/aws4_request
#' 3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2")
#'
#' @seealso \code{\link{signature_v4}}, \code{\link{signature_v4_auth}}
#' @export
string_to_sign <-
      function(
            algorithm = "AWS4-HMAC-SHA256",
            datetime, # format(Sys.time(),"%Y%m%dT%H%M%SZ", tz = "UTC")
            region,
            service,
            request_hash
      ) {
            paste(
                  algorithm,
                  datetime,
                  paste(
                        substring(datetime, 1, 8),
                        region,
                        service,
                        "aws4_request",
                        sep = "/"
                  ),
                  request_hash,
                  sep = "\n"
            )
      }

#' @title Signature Version 4
#' @description Generates AWS Signature Version 4
#' @template secret
#' @param date A character string containing a date in the form of \dQuote{YYMMDD}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.
#' @template region
#' @param service A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).
#' @param string_to_sign A character string containing the \dQuote{String To Sign}, possibly returned by \code{\link{string_to_sign}}.
#' @template verbose
#' @details This function generates an AWS Signature Version 4 for authorizing API requests from its pre-formatted components. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references
#'   \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html?refid=faq_cardl}{AWS General Reference: Signature Version 4 Signing Process}
#'
#'   \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{AWS General Reference: Examples of How to Derive a Version 4 Signing Key}
#'
#'   \href{https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html}{Amazon S3 API Reference: Authenticating Requests (AWS Signature Version 4)}
#' @seealso \code{\link{signature_v4_auth}}, \code{\link{signature_v2_auth}}, \code{\link{use_credentials}}
#' @examples
#' \dontrun{
#' # From AWS documentation
#' # http://docs.aws.amazon.com/general/latest/gr/signature-v4-test-suite.html
#' StringToSign <- "AWS4-HMAC-SHA256
#' 20110909T233600Z
#' 20110909/us-east-1/host/aws4_request
#' e25f777ba161a0f1baf778a87faf057187cf5987f17953320e3ca399feb5f00d"
#'
#' sig <-
#' signature_v4(secret = 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY',
#'              date = '20110909',
#'              region = 'us-east-1',
#'              service = 'host',
#'              string_to_sign = StringToSign)
#' identical(sig, "be7148d34ebccdc6423b19085378aa0bee970bdc61d144bd1a8c48c33079ab09")
#'
#' # http://docs.aws.amazon.com/general/latest/gr/sigv4-calculate-signature.html
#' StringToSign <- "AWS4-HMAC-SHA256
#' 20110909T233600Z
#' 20110909/us-east-1/iam/aws4_request
#' 3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2"
#'
#' sig <-
#' signature_v4(secret = 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY',
#'              date = '20110909',
#'              region = 'us-east-1',
#'              service = 'iam',
#'              string_to_sign = StringToSign)
#' identical(sig, "ced6826de92d2bdeed8f846f0bf508e8559e98e4b0199114b84c54174deb456c")
#' }
#' @importFrom digest digest hmac
#' @export
signature_v4 <-
  function(
    secret = NULL,
    date = format(Sys.time(), "%Y%m%d"),
    region = NULL,
    service,
    string_to_sign,
    verbose = getOption("verbose", FALSE)
  ) {
    kDate <- digest::hmac(paste0("AWS4", secret), date, "sha256", raw = TRUE)
    kRegion <- digest::hmac(kDate, region, "sha256", raw = TRUE)
    kService <- digest::hmac(kRegion, service, "sha256", raw = TRUE)
    kSigning <- digest::hmac(kService, "aws4_request", "sha256", raw = TRUE)
    signature <- digest::hmac(kSigning, string_to_sign, "sha256")
    return(signature)
  }

#' @rdname locate_credentials
#' @title Locate AWS Credentials
#' @description Locate AWS credentials from likely sources
#' @param key An AWS Access Key ID
#' @param secret An AWS Secret Access Key
#' @param session_token Optionally, an AWS Security Token Service (STS) temporary Session Token
#' @param region A character string containing the AWS region for the request. If missing, \dQuote{us-east-1} is assumed.
#' @param file A character string containing a path to a centralized \samp{.aws/credentials} file.
#' @param profile A character string specifying which profile to use from the file. By default, the profile named in \env{AWS_PROFILE} is used, otherwise the \dQuote{default} profile is used.
#' @param default_region A character string specifying a default string to use of no user-supplied value is found.
#' @param verbose A logical indicating whether to be verbose.
#' @details These functions locate values of AWS credentials (access key, secret access key, session token, and region) from likely sources. The order in which these are searched is as follows:
#' \enumerate{
#'   \item user-supplied values passed to the function
#'   \item environment variables, first checking for default credentials (\env{AWS_ACCESS_KEY_ID}, \env{AWS_SECRET_ACCESS_KEY}, \env{AWS_DEFAULT_REGION}, and \env{AWS_SESSION_TOKEN}); then for Web Identity Provider credentials (\env{AWS_ROLE_ARN} and \env{AWS_WEB_IDENTITY_TOKEN_FILE})
#'   \item an instance role (on the running ECS task from which this function is called) as identified by \code{\link[aws.ec2metadata]{metadata}}, if the aws.ec2metadata package is installed
#'   \item an IAM instance role (on the running EC2 instance from which this function is called) as identified by \code{\link[aws.ec2metadata]{metadata}}, if the aws.ec2metadata package is installed
#'   \item a profile in a local credentials dot file in the current working directory, using the profile specified by \env{AWS_PROFILE}
#'   \item the default profile in that local credentials file
#'   \item a profile in a global credentials dot file in a location set by \env{AWS_SHARED_CREDENTIALS_FILE} or defaulting typically to \file{~/.aws/credentials} (or another OS-specific location), using the profile specified by \env{AWS_PROFILE}
#'   \item the default profile in that global credentials file
#' }
#' 
#' If \env{AWS_ACCESS_KEY_ID} and \env{AWS_SECRET_ACCESS_KEY} environment variables are not present when the package is loaded, then \code{use_credentials} is invoked using the file specified in \env{AWS_SHARED_CREDENTIALS_FILE} (or another default location) and the profile specified in \env{AWS_PROFILE} (or, if missing, the \dQuote{default} profile).
#' 
#' To use this (and any cloudyr package) on AWS EC2 instances, users will also need to install the \href{https://cran.r-project.org/package=aws.ec2metadata}{aws.ec2metadata} package, which allows \code{locate_credentials} to know it is running in an instance and check for relevant values. If this package is not installed, instance metadata is not checked.
#' 
#' Because region is often handled slightly differently from credentials and is required for most requests (whereas some services allow anonymous requests without specifying credentials), the value of region is searched for in the same order as the above but lacking a value there fails safe with the following preference ranking of possible region values (regardless of location of other credentials):
#' \enumerate{
#'   \item a user-supplied value
#'   \item the \env{AWS_DEFAULT_REGION} environment variable
#'   \item (only on EC2 instances) a region declared in the instance metadata
#'   \item (if a credentials file is being used) the value specified therein
#'   \item the default value specified in \code{default_region} (i.e., \dQuote{us-east-1} - this can be overriden with the option \dQuote{cloudyr.aws.default_region})
#' }
#' 
#' As such, user-supplied values of \code{region} always trump any other value.
#' 
#' @seealso \code{\link{signature_v4}}, \code{\link{signature_v2_auth}}, \code{\link{use_credentials}}
#' @export
locate_credentials <-
function(
  key = NULL,
  secret = NULL,
  session_token = NULL,
  region = NULL,
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file()),
  profile = NULL,
  default_region = getOption("cloudyr.aws.default_region", "us-east-1"),
  verbose = getOption("verbose", FALSE)
) {

    credentials_feature_flag_on <- getOption("cloudyr.aws.prioritise_passed_profile", TRUE)
    
    if (isTRUE(verbose)) {
        message("Locating credentials")
    }
    
    # check for user-supplied values
    user_supplied_values <- check_for_user_supplied_values(key, secret, region, session_token, default_region, verbose)
    if (!is.null(user_supplied_values)){
      # early return
      return(user_supplied_values)
    }

    # Check for user-supplied profile
    if (credentials_feature_flag_on) {
      user_supplied_profile <- check_for_user_supplied_profile(profile, file, region, session_token, default_region, verbose)
      if (!is.null(user_supplied_profile)){
        # early return
        return(user_supplied_profile)
      }
    }

    # otherwise use environment variables if no user-supplied values
    env_vars <- check_for_env_vars(region, file, default_region, session_token, verbose)
    if (!is.null(env_vars)){
      # early return
      return(env_vars)
    }

    # lacking that, check for ECS metadata
    ecs_metadata <- check_ecs_metadata(region, default_region, verbose)
    if (!is.null(ecs_metadata)){
      # early return
      return(ecs_metadata)
    }

    # lacking that, check for EC2 metadata
    ec2_metadata <- check_ec2_metadata(region, default_region, verbose)
    if (!is.null(ec2_metadata)){
      # early return
      return(ec2_metadata)
    }

    # lastly, check for credentials file
    creds_file <- check_credentials_file(profile, file, region, default_region, verbose)
    if (!is.null(creds_file)){
      # early return
      return(creds_file)
    }
        
    # if that fails, no credentials can be found anywhere
    if (isTRUE(verbose)) {
        message("No user-supplied credentials, environment variables, instance metadata, or credentials file found!")
    }
    # find region, with fail safes
    region <- find_region_with_failsafe(region = region, default_region = default_region, verbose = verbose)
    
    # return identified values
    return(list(key = key, secret = secret, session_token = session_token, region = region))
}

check_ec2 <- function() {
    ec2 <- FALSE
    if (requireNamespace("aws.ec2metadata", quietly = TRUE)) {
        if (aws.ec2metadata::is_ec2()) {
            ec2 <- TRUE
        }
    }
    return(ec2)
}

check_ecs <- function() {
    ecs <- FALSE
    if (requireNamespace("aws.ec2metadata", quietly = TRUE)) {
        if (aws.ec2metadata::is_ecs()) {
            ecs <- TRUE
        }
    }
    return(ecs)
}

get_ec2_role <- function(role, verbose = getOption("verbose", FALSE)) {
    if (!requireNamespace("aws.ec2metadata", quietly = TRUE)) {
        return(NULL)
    }
    if (!isTRUE(aws.ec2metadata::is_ec2())) {
        return(NULL)
    }
    if (missing(role)) {
        role <- try(aws.ec2metadata::metadata$iam_role_names(), silent = TRUE)
        if (!length(role)) {
          if (isTRUE(verbose)) {
            warning("No IAM role profile available in instance metadata")
          }
          return(NULL)
        }
        if (isTRUE(verbose)) {
            message("Using EC2 Instance Metadata")
        }
    }
    # return role credentials as list
    out <- try(aws.ec2metadata::metadata$iam_role(role[1L]), silent = TRUE)
    if (inherits(out, "try-error")) {
        out <- NULL
    }
    return(out)
}

credentials_to_list <-
function(
  cred,
  region = NULL,
  default_region = getOption("cloudyr.aws.default_region", "us-east-1"),
  verbose = getOption("verbose", FALSE)
) {
    key <- NULL
    secret <- NULL
    session_token <- NULL

    if (!is.null(cred[["AWS_ACCESS_KEY_ID"]])) {
        key <- cred[["AWS_ACCESS_KEY_ID"]]
        vmsg(verbose, "Using value in credentials file for AWS Access Key ID")
    }
    if (!is.null(cred[["AWS_SECRET_ACCESS_KEY"]])) {
        secret <- cred[["AWS_SECRET_ACCESS_KEY"]]
        vmsg(verbose, "Using value in credentials file for AWS Secret Access Key")
    }
    if (!is.null(cred[["AWS_SESSION_TOKEN"]])) {
        session_token <- cred[["AWS_SESSION_TOKEN"]]
        vmsg(verbose, "Using value in credentials file for AWS Session Token")
    } else {
        session_token <- NULL
    }
    # now find region, with fail safes (including credentials file)
    if (!is_blank(region)||(identical(region, "")&&getOption("cloudyr.aws.allow_empty_region", FALSE))) {
        region <- region
        vmsg(verbose, "Using user-supplied value for AWS Region ('%s')", region)

    } else if (!is.null(cred[["AWS_DEFAULT_REGION"]]) && cred[["AWS_DEFAULT_REGION"]] != "") {
        region <- cred[["AWS_DEFAULT_REGION"]]
        vmsg(verbose, "Using value in credentials file for AWS Region ('%s')", region)
    } else {
        region <- Sys.getenv("AWS_DEFAULT_REGION")
        if (!is.null(region) && region != "") {
          vmsg(verbose, "Using Environment Variable 'AWS_DEFAULT_REGION' for AWS Region ('%s')", region)
        } else {
            region <- default_region
            vmsg(verbose, "Using default value for AWS Region ('%s')", region )
        }
    }
    
    # return values
    return(list(key = key, secret = secret, session_token = session_token, region = region))
}

find_region_with_failsafe <-
function(
  region = NULL,
  default_region = getOption("cloudyr.aws.default_region", "us-east-1"),
  verbose = getOption("verbose", FALSE),
  try_ec2 = FALSE
) {
  # if we have a valid argument, just return it
  if (!is_blank(region)||(identical(region,"")&&getOption("cloudyr.aws.allow_empty_region", FALSE))) {
    vmsg(verbose, "Using user-supplied value for AWS Region ('%s')", region)
    return(region)
  }
  
  # look in the environment
  region <- Sys.getenv("AWS_DEFAULT_REGION")
  if (!is_blank(region)) {
    vmsg(verbose, "Using Environment Variable 'AWS_DEFAULT_REGION' for AWS Region ('%s')", region)
    return(region)
  }
  
  # if we've been asked to, try the ec2 instance metadata
  # this isn't done by default as its expensive if not on EC2
  if (isTRUE(try_ec2)) {
    reg <- try(aws.ec2metadata::instance_document()$region, silent = TRUE)
    if (!inherits(reg, "try-error")&&!is_blank(reg)) {
      region <- reg
      vmsg(verbose, "Using EC2 Instance Metadata for AWS Region ('%s')", region)
      return(region)
    }
    
  }

  # otherwise, use the default region
  region <- default_region
  vmsg(verbose, "Using default value for AWS Region ('%s')", region)
  
  return(region)
}

check_for_user_supplied_values <- function(key, secret, region, session_token, default_region, verbose) {
  if (isTRUE(verbose)) {
      message("Checking for credentials in user-supplied values")
  }
  if (!is_blank(key) && !is_blank(secret)) {
      if (isTRUE(verbose)) {
          message("Using user-supplied value for AWS Access Key ID")
          message("Using user-supplied value for AWS Secret Access Key")
      }
      
      if (!is_blank(session_token)) {
          if (isTRUE(verbose)) {
              message("Using user-supplied value for AWS Session Token")
          }
      }
      # now find region, with fail safes
      region <- find_region_with_failsafe(region = region, default_region = default_region, verbose = verbose)
      
      return(list(key = key, secret = secret, session_token = session_token, region = region))
  } else{
    return(NULL)
  }
}

check_for_user_supplied_profile <- function(profile, file, region, session_token, default_region, verbose) {
    if (is.null(profile)) {
        return(NULL)
    }
    if (isTRUE(verbose)) {
        message("Checking for credentials from user-defined profile")
    }
    profile_creds <- check_credentials_file(profile, file, region, default_region, verbose, require_profile = TRUE)
    if (!is.null(profile_creds)) {
        return(profile_creds)
    } 
    # If we can't find the profile the user gave us, bail out
    if (!is.null(profile)) {
        stop(paste0("Could not find profile matching '", profile, "'. Check that the profile you passed in matches your credentials file."))
    }
    return(NULL)
}


check_for_env_vars <- function(region, file, default_region, session_token, verbose) {

    if (isTRUE(verbose)) {
      message("Checking for credentials in Environment Variables")
    }

    # try to use environment variables if no user-supplied values
    # grab environment variables
    env <- list(key = Sys.getenv("AWS_ACCESS_KEY_ID"),
                secret = Sys.getenv("AWS_SECRET_ACCESS_KEY"),
                profile = Sys.getenv("AWS_PROFILE"),
                session_token = Sys.getenv("AWS_SESSION_TOKEN"),
                region = Sys.getenv("AWS_DEFAULT_REGION"),
                arn=Sys.getenv("AWS_ROLE_ARN"),
                token_file=Sys.getenv("AWS_WEB_IDENTITY_TOKEN_FILE"))

    if (!is_blank(env$session_token)) {
        session_token <- env$session_token
        if (isTRUE(verbose)) {
            message("Using Environment Variable 'AWS_SESSION_TOKEN' for AWS Session Token")
        }
    } else {
        if (!is_blank(session_token)) {
            session_token <- session_token
            if (isTRUE(verbose)) {
                message("Using user-supplied value for AWS Session Token")
            }
        }
    }
    
    if (!is_blank(env$key)&&!is_blank(env$secret)) {
        if (isTRUE(verbose)) {
            message("Using Environment Variable 'AWS_ACCESS_KEY_ID' for AWS Access Key ID")
            message("Using Environment Variable 'AWS_SECRET_ACCESS_KEY' for AWS Secret Access Key")
        }
        key <- env$key
        secret <- env$secret

        # now find region, with fail safes
        region <- find_region_with_failsafe(region = region, default_region = default_region, verbose = verbose)
        
        # early return
        return(list(key = key, secret = secret, session_token = session_token, region = region))

    } else if (!is_blank(env$arn) && !is_blank(env$token_file)){  # Web Identity Provider
        if (isTRUE(verbose)) {
            message("Using Environment Variables 'AWS_WEB_IDENTITY_TOKEN_FILE' and `AWS_ROLE_ARN`")
            message("to assume role with Web Identity Provider")
        }
      response <- assume_role_with_web_identity(env$arn, env$token_file)

      creds <- response$AssumeRoleWithWebIdentityResponse$AssumeRoleWithWebIdentityResult$Credentials
      key <- creds$AccessKeyId
      secret <- creds$SecretAccessKey
      session_token <- creds$SessionToken

      region <- find_region_with_failsafe(region = region, default_region = default_region, verbose = verbose)

      return(list(key = key, secret = secret, session_token = session_token, region = region))
    } else if (!is_blank(env$profile)) {
        return(check_credentials_file(env$profile, file, region, default_region, verbose))
    }
    return(NULL)
}

check_ecs_metadata <- function(region, default_region, verbose){
    if (isTRUE(check_ecs())) {
      if (isTRUE(verbose)) {
        message("Checking for credentials in ECS Instance Metadata")
      }
      ecs_meta <- aws.ec2metadata::metadata$ecs_task_role()
      if (!is.null(ecs_meta)) {
        if (isTRUE(verbose)) {
          message(sprintf("Using ECS Instance Metadata for AWS Credentials (IAM Role: %s)", ecs_meta$RoleArn))
        }
        
        key <- ecs_meta[["AccessKeyId"]]
        secret <- ecs_meta[["SecretAccessKey"]]
        session_token <- ecs_meta[["Token"]]
        
        # now find region, with fail safes
        region <- find_region_with_failsafe(region = region,
                                            default_region = default_region,
                                            verbose = verbose,
                                            try_ec2 = TRUE)
        # early return
        return(list(key = key,
                    secret = secret,
                    session_token = session_token,
                    region = region)
               )
      }
    }
    return(NULL)
}

check_ec2_metadata <- function(region, default_region, verbose){
    if (isTRUE(check_ec2())) {
        if (isTRUE(verbose)) {
            message("Checking for credentials in EC2 Instance Metadata")
        }
        role <- try(get_ec2_role(verbose = verbose), silent = TRUE)
        if (!is.null(role)&&!inherits(role, "try-error")) {
          if (isTRUE(verbose)) {
            message("Using EC2 Instance Metadata for credentials")
          }
          key <- role[["AccessKeyId"]]
          secret <- role[["SecretAccessKey"]]
          session_token <- role[["Token"]]
          
          # now find region, with fail safes
          region <- find_region_with_failsafe(region = region,
                                              default_region = default_region,
                                              verbose = verbose,
                                              try_ec2 = TRUE)
          # early return
          return(list(key = key, secret = secret, session_token = session_token, region = region))
        }
    }
    return(NULL)
}

check_credentials_file <- function(profile, file, region, default_region, verbose, require_profile=FALSE) {
    vmsg(verbose, "Searching for credentials file(s)")

    files_to_check <- c(
      file.path(".aws", "credentials"),
      file,
      default_credentials_file()
    )
    files_to_check <- unique(files_to_check)
    
    for (f in files_to_check) {
      vmsg(verbose, sprintf("Attempting to read credentials from '%s'", f))
      
      creds <- read_profile_from_credentials_file(profile, f, region, default_region, verbose, require_profile)
      if (!is.null(creds)) {
        break
      }
    }
    
    return(creds)
}

read_profile_from_credentials_file <- function(profile, file, region, default_region, verbose, require_profile=FALSE) {
  if (!file.exists(file)) return(NULL)
  
  cred <- read_credentials(file)
  
  if (missing(profile)||is.null(profile)) {
    profile <- "default"
  }
  vmsg(verbose, sprintf("Attempting to find profile '%s'", profile))
  
  if (profile %in% names(cred)) {
    cred <- cred[[profile]]
    vmsg(verbose, sprintf("Using profile '%s' from credentials files from '%s'", profile, file))
  } else {
    if (isTRUE(require_profile)) {
      return(NULL)
    } else if ("default" %in% names(cred)) {
      cred <- cred[["default"]]
      vmsg(verbose, sprintf("Using profile '%s' from credentials files from '%s'", "default", file))
    } else {
      return(NULL)
    }
  }
  
  return(credentials_to_list(cred, region = region, default_region = default_region, verbose = verbose))
}

context("Tests locate_credentials()")

if (file.exists(".aws/credentials")) {
    
    test_that("locate_credentials() returns envvar values when environment variables and credentials file present", {
        
        skip_on_cran()
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION", "AWS_PROFILE"))
        
        # set environment variables
        Sys.setenv("AWS_ACCESS_KEY_ID" = "foo-key")
        Sys.setenv("AWS_SECRET_ACCESS_KEY" = "foo-secret")
        Sys.setenv("AWS_SESSION_TOKEN" = "foo-token")
        Sys.setenv("AWS_DEFAULT_REGION" = "foo-region")
        Sys.unsetenv("AWS_PROFILE")
        
        # tests
        cred <- locate_credentials()
        expect_equal(cred[["key"]], "foo-key")
        expect_equal(cred[["secret"]], "foo-secret")
        expect_equal(cred[["session_token"]], "foo-token")
        expect_equal(cred[["region"]], "foo-region")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })
    
    test_that("locate_credentials() returns non-default values if requested when environment variables and credentials file present", {
        
        skip_on_cran()
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
        
        # set environment variables
        Sys.setenv("AWS_ACCESS_KEY_ID" = "foo-key")
        Sys.setenv("AWS_SECRET_ACCESS_KEY" = "foo-secret")
        Sys.setenv("AWS_SESSION_TOKEN" = "foo-token")
        Sys.setenv("AWS_DEFAULT_REGION" = "foo-region")
        
        # tests
        cred <- locate_credentials(key = "foo1", secret = "foo2", session_token = "foo3", region = "foo4")
        expect_equal(cred[["key"]], "foo1", label = "locate_credentials(key = 'foo1')")
        expect_equal(cred[["secret"]], "foo2", label = "locate_credentials(secret = 'foo2')")
        expect_equal(cred[["session_token"]], "foo3", label = "locate_credentials(session_token = 'foo3')")
        expect_equal(cred[["region"]], "foo4", label = "locate_credentials(region = 'foo4')")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })
    
    test_that("locate_credentials() returns credentials file values when environment variables missing and credentials file present", {
        
        skip_on_cran()
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
        
        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        
        # tests
        cred <- locate_credentials()
        expect_equal(cred[["key"]], "ACCESS_KEY")
        expect_equal(cred[["secret"]], "SECRET_KEY")
        expect_equal(cred[["session_token"]], "TOKEN")
        expect_equal(cred[["region"]], "us-east-1")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })
    
    test_that("locate_credentials() returns non-default values if requested, when environment variables missing and credentials file absent", {
        
        skip_on_cran()
        
        # move credentials file
        tmp <- tempfile(tmpdir = ".aws")
        file.rename(".aws/credentials", tmp)
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
        
        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        
        # tests
        cred <- locate_credentials(key = "foo1", secret = "foo2", session_token = "foo3", region = "foo4")
        expect_equal(cred[["key"]], "foo1", label = "locate_credentials(key = 'foo1')")
        expect_equal(cred[["secret"]], "foo2", label = "locate_credentials(secret = 'foo2')")
        expect_equal(cred[["session_token"]], "foo3", label = "locate_credentials(session_token = 'foo3')")
        expect_equal(cred[["region"]], "foo4", label = "locate_credentials(region = 'foo4')")
        
        # restore credentials file
        file.rename(tmp, ".aws/credentials")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })
    
    test_that("locate_credentials() returns NULLs when environment variables missing and credentials file absent", {
        
        skip_on_cran()
        
        # move credentials file
        tmp <- tempfile(tmpdir = ".aws")
        file.rename(".aws/credentials", tmp)
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
        
        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        
        # tests
        cred <- locate_credentials()
        expect_null(cred[["key"]])
        expect_null(cred[["secret"]])
        expect_null(cred[["session_token"]])
        expect_equal(cred[["region"]], "us-east-1")
        
        # restore credentials file
        file.rename(tmp, ".aws/credentials")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })

    test_that("locate_credentials() uses options default for region", {

        skip_on_cran()

        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))

        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        cloudyr_region <- getOption("cloudyr.aws.default_region")
        options(cloudyr.aws.default_region = "eu-west-1")

        # tests
        cred <- locate_credentials()
        expect_equal(cred[["key"]], "ACCESS_KEY")
        expect_equal(cred[["secret"]], "SECRET_KEY")
        expect_equal(cred[["session_token"]], "TOKEN")
        expect_equal(cred[["region"]], "eu-west-1")

        # restore environment variables
        do.call("Sys.setenv", as.list(e))
        options(cloudyr.aws.default_region = cloudyr_region)
    })

    test_that("locate_credentials() allows empty region with allow_empty_set", {

        skip_on_cran()

        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))

        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        cloudyr_region <- getOption("cloudyr.aws.allow_empty_region")
        options(cloudyr.aws.allow_empty_region = TRUE)

        # tests
        cred <- locate_credentials(region="")
        expect_equal(cred[["key"]], "ACCESS_KEY")
        expect_equal(cred[["secret"]], "SECRET_KEY")
        expect_equal(cred[["session_token"]], "TOKEN")
        expect_equal(cred[["region"]], "")

        # restore environment variables
        do.call("Sys.setenv", as.list(e))
        options(cloudyr.aws.allow_empty_region = cloudyr_region)
    })
    
    test_that("locate_credentials() works with allow_empty_set", {
        
        skip_on_cran()
        
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION"))
        
        # unset environment variables
        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        Sys.unsetenv("AWS_SESSION_TOKEN")
        Sys.unsetenv("AWS_DEFAULT_REGION")
        cloudyr_region <- getOption("cloudyr.aws.allow_empty_region")
        options(cloudyr.aws.allow_empty_region = TRUE)
        
        # tests
        cred <- locate_credentials(region=NULL)
        expect_equal(cred[["key"]], "ACCESS_KEY")
        expect_equal(cred[["secret"]], "SECRET_KEY")
        expect_equal(cred[["session_token"]], "TOKEN")
        expect_equal(cred[["region"]], "us-east-1")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
        options(cloudyr.aws.allow_empty_region = cloudyr_region)
    })

    test_that("locate_credentials() prioritizes profile argument when conflicting env-vars are set", {

        skip_on_cran()

        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID",
                          "AWS_SECRET_ACCESS_KEY",
                          "AWS_SESSION_TOKEN",
                          "AWS_DEFAULT_REGION"
                          ))
        
        # set environment variables
        Sys.setenv("AWS_ACCESS_KEY_ID" = "foo-key")
        Sys.setenv("AWS_SECRET_ACCESS_KEY" = "foo-secret")
        Sys.setenv("AWS_SESSION_TOKEN" = "foo-token")
        Sys.setenv("AWS_DEFAULT_REGION" = "foo-region")

        # tests
        cred <- locate_credentials(profile = 'Alice')
        expect_equal(cred[["key"]], "Alice_access_key_ID")
        expect_equal(cred[["secret"]], "Alice_secret_access_key")
        
        # restore environment variables
        do.call("Sys.setenv", as.list(e))
    })

    test_that("locate_credentials() prioritizes env var keys if there is a conflicting env profile var", {
        skip_on_cran()
        # save environment variables
        e <- Sys.getenv(c("AWS_ACCESS_KEY_ID",
                          "AWS_SECRET_ACCESS_KEY",
                          "AWS_SESSION_TOKEN",
                          "AWS_PROFILE",
                          "AWS_DEFAULT_REGION"
                          ))

        # set environment variables
        Sys.setenv("AWS_ACCESS_KEY_ID" = "foo-key")
        Sys.setenv("AWS_SECRET_ACCESS_KEY" = "foo-secret")
        Sys.setenv("AWS_PROFILE" = "Alice")
        Sys.setenv("AWS_SESSION_TOKEN" = "foo-token")
        Sys.setenv("AWS_DEFAULT_REGION" = "foo-region")

        # tests
        cred <- locate_credentials()
        expect_equal(cred[["key"]], "foo-key")
        expect_equal(cred[["secret"]], "foo-secret")

        Sys.unsetenv("AWS_ACCESS_KEY_ID")
        Sys.unsetenv("AWS_SECRET_ACCESS_KEY")
        cred <- locate_credentials()
        expect_equal(cred[["key"]], "Alice_access_key_ID")
        expect_equal(cred[["secret"]], "Alice_secret_access_key")

        do.call("Sys.setenv", as.list(e))
    })

    test_that("locate_credentials() errors on unknown profile", {
        expect_error(locate_credentials(profile = "buzz"), 
                     "Could not find profile matching" )
    })
}

test_that("locate_credentials() is noisy with verbose = TRUE", {
    
    skip_on_cran()
    
    # save environment variables
    e <- Sys.getenv(c("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN", "AWS_DEFAULT_REGION", "AWS_PROFILE"))
    
    # set environment variables
    Sys.setenv("AWS_ACCESS_KEY_ID" = "foo-key")
    Sys.setenv("AWS_SECRET_ACCESS_KEY" = "foo-secret")
    Sys.setenv("AWS_SESSION_TOKEN" = "foo-token")
    Sys.setenv("AWS_DEFAULT_REGION" = "foo-region")
    Sys.unsetenv("AWS_PROFILE")
    
    # tests
    expect_message(locate_credentials(verbose = TRUE), "Locating credentials")
    
    # restore environment variables
    do.call("Sys.setenv", as.list(e))
})

#' @title Signature Version 4
#' @description AWS Signature Version 4 for use in query or header authorization
#' @param datetime A character string containing a datetime in the form of \dQuote{YYYYMMDDTHHMMSSZ}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.
#' @template region
#' @param service A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).
#' @param verb A character string containing the HTTP verb being used in the request.
#' @param action A character string containing the API endpoint used in the request.
#' @param query_args A named list of character strings containing the query string values (if any) used in the API request, passed to \code{\link{canonical_request}}.
#' @param canonical_headers A named list of character strings containing the headers used in the request.
#' @param request_body The body of the HTTP request.
#' @param signed_body Should the body be signed
#' @template key
#' @template secret
#' @param session_token Optionally, an AWS Security Token Service (STS) temporary Session Token. This is added automatically as a header to \code{canonical_headers}. See \code{\link{locate_credentials}}.
#' @param query A logical. Currently ignored.
#' @param algorithm A character string containing the hashing algorithm used in the request. Should only be \dQuote{SHA256}.
#' @template force_credentials
#' @template verbose
#' @details This function generates an AWS Signature Version 4 for authorizing API requests.
#' @return A list of class \dQuote{aws_signature_v4}, containing the information needed to sign an AWS API request using either query string authentication or request header authentication. Specifically, the list contains:
#'
#'     \item{Algorithm}{A character string containing the hashing algorithm used during the signing process (default is SHA256).}
#'     \item{Credential}{A character string containing an identifying credential \dQuote{scoped} to the region, date, and service of the request.}
#'     \item{Date}{A character string containing a YYYYMMDD-formatted date.}
#'     \item{SignedHeaders}{A character string containing a semicolon-separated listing of request headers used in the signature.}
#'     \item{Body}{The value passed to \code{request_body}.}
#'     \item{BodyHash}{A character string containing a SHA256 hash of the request body.}
#'     \item{Verb}{The value passed to \code{verb}.}
#'     \item{Query}{The value passed to \code{query_args}.}
#'     \item{Service}{The value passed to \code{service}.}
#'     \item{Action}{The value passed to \code{action}.}
#'     \item{CanonicalRequest}{A character string containing the canonical request.}
#'     \item{StringToSign}{A character string containing the string to sign for the request.}
#'     \item{Signature}{A character string containing a request signature hash.}
#'     \item{SignatureHeader}{A character string containing a complete Authorization header value.}
#'     \item{AccessKeyId}{A character string containing the access key id identified by \code{\link{locate_credentials}}.}
#'     \item{SecretAccessKey}{A character string containing the secret access key identified by \code{\link{locate_credentials}}.}
#'     \item{SessionToken}{A character string containing the session token identified by \code{\link{locate_credentials}}.}
#'     \item{Region}{A character string containing the region identified by \code{\link{locate_credentials}}.}
#'
#' These values can either be used as query parameters in a REST-style API request, or as request headers. If authentication is supplied via query string parameters, the query string should include the following:
#'
#' Action=\code{action}
#' &X-Amz-Algorithm=\code{Algorithm}
#' &X-Amz-Credential=\code{URLencode(Credentials)}
#' &X-Amz-Date=\code{Date}
#' &X-Amz-Expires=\code{timeout}
#' &X-Amz-SignedHeaders=\code{SignedHeaders}
#'
#' where \code{action} is the API endpoint being called and \code{timeout} is a numeric value indicating when the request should expire.
#'
#' If signing a request using header-based authentication, the \dQuote{Authorization} header in the request should be included with the request that looks as follows:
#'
#' Authorization: \code{Algorithm} Credential=\code{Credential}, SignedHeaders=\code{SignedHeaders}, Signature=\code{Signature}
#'
#' This is the value printed by default for all objects of class \dQuote{aws_signature_v4}.
#' @author Thomas J. Leeper <thosjleeper@gmail.com>
#' @references
#' \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html?refid=faq_card}{AWS General Reference: Signature Version 4 Signing Process}
#'
#' \href{https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html}{Amazon S3 API Reference: Authenticating Requests (AWS Signature Version 4)}
#'
#' \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Add the Signing Information to the Request}
#' @seealso \code{\link{signature_v2_auth}}, \code{\link{locate_credentials}}
#' @export
signature_v4_auth <-
    function(
        datetime = format(Sys.time(), "%Y%m%dT%H%M%SZ", tz = "UTC"),
        region = NULL,
        service,
        verb,
        action,
        query_args = list(),
        canonical_headers, # named list
        request_body,
        signed_body = FALSE,
        key = NULL,
        secret = NULL,
        session_token = NULL,
        query = FALSE,
        algorithm = "AWS4-HMAC-SHA256",
        force_credentials = FALSE,
        verbose = getOption("verbose", FALSE)
    ) {
        if (isTRUE(force_credentials)) {
            if (isTRUE(verbose)) {
                if (!is.null(key)) {
                    message("Using user-supplied value for AWS Access Key ID")
                }
                if (!is.null(secret)) {
                    message(
                        "Using user-supplied value for AWS Secret Access Key"
                    )
                }
                if (!is.null(session_token)) {
                    message(
                        "Using user-supplied value for AWS Secret Access Key"
                    )
                }
                if (!is.null(region)) {
                    message(sprintf(
                        "Using user-supplied value for AWS Region ('%s')",
                        region
                    ))
                }
            }
        } else {
            credentials <- locate_credentials(
                key = key,
                secret = secret,
                session_token = session_token,
                region = region,
                verbose = verbose
            )
            key <- credentials[["key"]]
            secret <- credentials[["secret"]]
            session_token <- credentials[["session_token"]]
            region <- credentials[["region"]]
        }

        date <- substring(datetime, 1, 8)

        if (isTRUE(query)) {
            # handle query-based authorizations, by including relevant parameters
        }

        # Canonical Request
        if (!is.null(session_token) && session_token != "") {
            if (!missing(canonical_headers)) {
                canonical_headers <- c(
                    canonical_headers,
                    list("X-Amz-Security-Token" = session_token)
                )
            } else {
                canonical_headers <- list(
                    "X-Amz-Security-Token" = session_token
                )
            }
        }
        R <- canonical_request(
            verb = verb,
            canonical_uri = action,
            query_args = query_args,
            canonical_headers = canonical_headers,
            request_body = request_body,
            signed_body = signed_body
        )

        # String To Sign
        S <- string_to_sign(
            algorithm = algorithm,
            datetime = datetime,
            region = region,
            service = service,
            request_hash = R$hash
        )

        # Signature
        V4 <- signature_v4(
            secret = secret,
            date = date,
            region = region,
            service = service,
            string_to_sign = S,
            verbose = verbose
        )

        # return list
        credential <- paste(
            key,
            date,
            region,
            service,
            "aws4_request",
            sep = "/"
        )
        sigheader <- paste(
            algorithm,
            paste(
                paste0("Credential=", credential),
                paste0("SignedHeaders=", R$headers),
                paste0("Signature=", V4),
                sep = ","
            )
        )
        structure(
            list(
                Algorithm = algorithm,
                Credential = credential,
                Date = date,
                SignedHeaders = R$headers,
                Body = request_body,
                BodyHash = R$body,
                Verb = verb,
                Query = query_args,
                Service = service,
                Action = action,
                CanonicalRequest = R$canonical,
                StringToSign = S,
                Signature = V4,
                SignatureHeader = sigheader,
                AccessKeyId = key,
                SecretAccessKey = secret,
                SessionToken = session_token,
                Region = region
            ),
            class = "aws_signature_v4"
        )
    }

test_that("A GET request is correctly signed", {
  key <- "AKIAIOSFODNN7EXAMPLE"
  secret <- "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
  region <- "us-east-1"
  
  
  
  signed <- signature_v4_auth(
    datetime = "20130524T000000Z",
    service = "s3",
    verb = "GET",
    action = "/test.txt",
    canonical_headers = list(
      "Host"= "examplebucket.s3.amazonaws.com",
      "Range"= "bytes=0-9",
      "X-AMZ-Date"="20130524T000000Z"
    ),
    request_body = "",
    signed_body=TRUE,
    key = key,
    secret= secret,
    region = region
  )
  
  expect_identical(signed$SignatureHeader, "AWS4-HMAC-SHA256 Credential=AKIAIOSFODNN7EXAMPLE/20130524/us-east-1/s3/aws4_request,SignedHeaders=host;range;x-amz-content-sha256;x-amz-date,Signature=f0e8bdb87c964420e857bd35b5d6ed310bd44f0170aba48dd91039c6036bdb41")
})

#' @rdname assume_role_with_web_identity
#' @title Assume Role with AWS Web Identity
#' @description Assume a role from a provided Web Identity Token and ARN using AWS Secure Token Service (STS).
#' @param role_arn A character string containing the AWS Role Amazon Resource Name (ARN). This specifies the permissions you have to access other AWS services.
#' @param token_file A character string containing a path to a Web Identity Token file.
#' @param base_url The AWS STS endpoint to use to retrieve your credentials from.
#' @param session_name A character string optionally specifying the name.
#' @param duration The expiry time on the retrieved credentials. 
#' @param version The AWS STS specification version to use.
#' @param verbose A logical indicating whether to be verbose.
#' @export
assume_role_with_web_identity <- function(
  role_arn, 
  token_file, 
  base_url=Sys.getenv("AWS_STS_ENDPOINT", "https://sts.amazonaws.com"), 
  session_name=NULL, 
  duration=3600,
  version="2011-06-15",
  verbose = getOption("verbose", FALSE)
){
  if (is.null(session_name)) {
    # strip resource ID from arn and use as default session name
    # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html
    session_name <- gsub("/", "-", utils::tail(strsplit(role_arn, ":")[[1]], 1))
  }
  
  token <- readChar(token_file, file.info(token_file)$size)

  query_params <- list(
    Action="AssumeRoleWithWebIdentity",
    DurationSeconds=duration,
    RoleArn=role_arn,
    RoleSessionName=session_name,
    WebIdentityToken=token,
    Version=version
  )
  query_params_names  <- curl::curl_escape(names(query_params))
  query_params_values <- lapply(query_params, curl::curl_escape)
  query_str  <- paste0(query_params_names, "=", query_params_values, collapse = "&")
  query_url <- paste0(base_url, "/?", query_str)

  handle <- curl::new_handle()  # need to accept json headers
  curl::handle_setheaders(handle, "accept" = "application/json")

  response <- curl::curl_fetch_memory(query_url, handle = handle)
  content <- jsonlite::fromJSON(rawToChar(response$content))
  
  if (response$status_code == 200) {
    if (isTRUE(verbose)) {
      message("Successfully fetched token from web identiy provider.")
    }
    return(content)
  } else {
    stop("Failed to assume role.")
  }
}

test_that("NULL is blank", {
  expect_true(is_blank(NULL))
})

test_that("empty string is blank", {
  expect_true(is_blank(""))
})

test_that("Non-empty string is not blank", {
  expect_false(is_blank("foo"))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signature_v4.R
\name{signature_v4}
\alias{signature_v4}
\title{Signature Version 4}
\usage{
signature_v4(
  secret = NULL,
  date = format(Sys.time(), "\%Y\%m\%d"),
  region = NULL,
  service,
  string_to_sign,
  verbose = getOption("verbose", FALSE)
)
}
\arguments{
\item{secret}{An AWS Secret Access Key. If \code{NULL}, it is retrieved using \code{\link{locate_credentials}}.}

\item{date}{A character string containing a date in the form of \dQuote{YYMMDD}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.}

\item{region}{A character string containing the AWS region for the request. If missing, \dQuote{us-east-1} is assumed.}

\item{service}{A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).}

\item{string_to_sign}{A character string containing the \dQuote{String To Sign}, possibly returned by \code{\link{string_to_sign}}.}

\item{verbose}{A logical indicating whether to be verbose.}
}
\description{
Generates AWS Signature Version 4
}
\details{
This function generates an AWS Signature Version 4 for authorizing API requests from its pre-formatted components. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
}
\examples{
\dontrun{
# From AWS documentation
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-test-suite.html
StringToSign <- "AWS4-HMAC-SHA256
20110909T233600Z
20110909/us-east-1/host/aws4_request
e25f777ba161a0f1baf778a87faf057187cf5987f17953320e3ca399feb5f00d"

sig <-
signature_v4(secret = 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY',
             date = '20110909',
             region = 'us-east-1',
             service = 'host',
             string_to_sign = StringToSign)
identical(sig, "be7148d34ebccdc6423b19085378aa0bee970bdc61d144bd1a8c48c33079ab09")

# http://docs.aws.amazon.com/general/latest/gr/sigv4-calculate-signature.html
StringToSign <- "AWS4-HMAC-SHA256
20110909T233600Z
20110909/us-east-1/iam/aws4_request
3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2"

sig <-
signature_v4(secret = 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY',
             date = '20110909',
             region = 'us-east-1',
             service = 'iam',
             string_to_sign = StringToSign)
identical(sig, "ced6826de92d2bdeed8f846f0bf508e8559e98e4b0199114b84c54174deb456c")
}
}
\references{
\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html?refid=faq_cardl}{AWS General Reference: Signature Version 4 Signing Process}

  \href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{AWS General Reference: Examples of How to Derive a Version 4 Signing Key}

  \href{https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html}{Amazon S3 API Reference: Authenticating Requests (AWS Signature Version 4)}
}
\seealso{
\code{\link{signature_v4_auth}}, \code{\link{signature_v2_auth}}, \code{\link{use_credentials}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_to_sign.R
\name{string_to_sign}
\alias{string_to_sign}
\title{Construct a String To Sign}
\usage{
string_to_sign(
  algorithm = "AWS4-HMAC-SHA256",
  datetime,
  region,
  service,
  request_hash
)
}
\arguments{
\item{algorithm}{A character string containing the hashing algorithm used in signing process. Should only be \dQuote{AWS4-HMAC-SHA256}.}

\item{datetime}{A character string containing a UTC date in the form of \dQuote{YYYYMMDDTHHMMSSZ}.}

\item{region}{A character string containing the AWS region for the request.}

\item{service}{A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).}

\item{request_hash}{A character string containing the hash of the canonical request, perhaps as returned by \code{\link{canonical_request}}.}
}
\description{
Construct a String to Sign from request elements
}
\details{
This is a mostly internal function that creates a \dQuote{String To Sign}, which is part of the Signature Version 4. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
}
\examples{
# From AWS documentation
rh <- "3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2"
sts <-
string_to_sign(datetime = "20110909T233600Z",
               region = "us-east-1",
               service = "iam",
               request_hash = rh)
identical(sts, "AWS4-HMAC-SHA256
20110909T233600Z
20110909/us-east-1/iam/aws4_request
3511de7e95d28ecd39e9513b642aee07e54f4941150d8df8bf94b328ef7e55e2")

}
\references{
\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Create a String to Sign for Signature Version 4}
}
\seealso{
\code{\link{signature_v4}}, \code{\link{signature_v4_auth}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws.signature-package.R
\docType{package}
\name{aws.signature-package}
\alias{aws.signature}
\alias{aws.signature-package}
\title{Amazon Web Services Request Signatures}
\description{
Generates Amazon Web Services (AWS) request signatures for RESTful APIs.
}
\details{
This package contains functions mostly intended for developers to use in
building API client packages for Amazon Web Services APIs.

The main function of interest is [signature_v4_auth()], which wraps the
other internal functions and returns a named list of elements to be used in
authenticating an API request using AWS Signature Version 4. Another
function, [signature_v2_auth()], implements the older, mostly deprecated
Version 2 algorithm.

Recent versions of the package (>= 0.2.8) identify credentials by walking
through a tree of possible sources of values (described in
[locate_credentials()]), with optional verbosity, in a manner similar to the
Python boto3 library.

A lower-level function that may be of use to end users is
[use_credentials()], which sets the environment variables used by this
package based upon values specified in a `.aws/credentials` file. That
function is called by default during package load, if no environment
variables are set.

To use this (and any cloudyr package) on AWS EC2 instances or ECS tasks,
users will also need to install the **aws.ec2metadata** package, which
allows [locate_credentials()] to know it is running in an instance and check
for relevant values.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/cloudyr/aws.signature}
  \item Report bugs at \url{https://github.com/cloudyr/aws.signature/issues}
}

}
\author{
\strong{Maintainer}: Gonzalo Diaz \email{gonzalomartindiaz22@gmail.com} (\href{https://orcid.org/0009-0003-9056-4896}{ORCID})

Authors:
\itemize{
  \item Gonzalo Diaz \email{gonzalomartindiaz22@gmail.com} (\href{https://orcid.org/0009-0003-9056-4896}{ORCID})
  \item Thomas J. Leeper \email{thosjleeper@gmail.com} (\href{https://orcid.org/0000-0003-4097-6326}{ORCID})
  \item Jonathan Stott \email{jonathan.stott@magairports.com}
}

Other contributors:
\itemize{
  \item Mike Kaminsky \email{kaminsky.michael@gmail.com} [contributor]
  \item Mark Douthwaite \email{mark.douthwaite@peak.ai} [contributor]
  \item Jason Gofford \email{jason.gofford@peak.ai} [contributor]
  \item Luke Dyer \email{luke.dyer@peak.ai} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/canonical_request.R
\name{canonical_request}
\alias{canonical_request}
\title{Construct a Canonical Request}
\usage{
canonical_request(
  verb,
  canonical_uri = "",
  query_args = list(),
  canonical_headers,
  request_body = "",
  signed_body = FALSE
)
}
\arguments{
\item{verb}{A character string containing the HTTP verb being used in the request.}

\item{canonical_uri}{A character string containing the \dQuote{canonical URI}, meaning the contents of the API request URI excluding the host and the query parameters.}

\item{query_args}{A named list of character strings containing the query string values (if any) used in the API request.}

\item{canonical_headers}{A named list of character strings containing the headers used in the request.}

\item{request_body}{The body of the HTTP request, or a filename. If a filename, hashing is performed on the file without reading it into memory.}

\item{signed_body}{Sign the body request and add the correct header (x-amz-content-sha256) to the list of headers}
}
\value{
A list containing
}
\description{
Construct a Canonical Request from request elements
}
\details{
This function creates a \dQuote{Canonical Request}, which is part of the Signature Version 4. Users probably only need to use the \code{\link{signature_v4_auth}} function to generate signatures.
}
\examples{
# From AWS documentation
# https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html
fromDocs <- "POST
/

content-type:application/x-www-form-urlencoded; charset=utf-8
host:iam.amazonaws.com
x-amz-date:20110909T233600Z

content-type;host;x-amz-date
b6359072c78d70ebee1e81adcbab4f01bf2c23245fa365ef83fe8f1f955085e2"

hdrs <- list(`Content-Type` = "application/x-www-form-urlencoded; charset=utf-8",
             Host = "iam.amazonaws.com",
             `x-amz-date` = "20110909T233600Z")
r <- canonical_request(verb = "POST",
                       canonical_uri = "/",
                       query_args = list(),
                       canonical_headers = hdrs,
                       request_body = "Action=ListUsers&Version=2010-05-08")

identical(fromDocs, r$canonical)
}
\references{
\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Create a Canonical Request For Signature Version 4}
}
\seealso{
\code{\link{signature_v4}}, \code{\link{signature_v4_auth}}

\code{link{signature_v4_aut}}, \code{\link{string_to_sign}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_credentials.R
\name{read_credentials}
\alias{read_credentials}
\alias{use_credentials}
\alias{default_credentials_file}
\title{Use Credentials from .aws/credentials File}
\usage{
read_credentials(
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file())
)

use_credentials(
  profile = Sys.getenv("AWS_PROFILE", "default"),
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file())
)

default_credentials_file()
}
\arguments{
\item{file}{A character string containing a path to a \samp{.aws/credentials} file. By default, the standard/centralized file given by \env{AWS_SHARED_CREDENTIALS_FILE} is used, otherwise an assumed default location is assumed. For \code{use_credentials}, this can also be an object of class \dQuote{aws_credentials} (as returned by \code{use_credentials}).}

\item{profile}{A character string specifying which profile to use from the file. By default, the \dQuote{default} profile is used.}
}
\description{
Use a profile from a \samp{.aws/credentials} file
}
\details{
\code{read_credentials} reads and parses a \samp{.aws/credentials} file into an object of class \dQuote{aws_credentials}.

\code{use_credentials} uses credentials from a profile stored in a credentials file to set the environment variables used by this package. It is called by default during package load if the \env{AWS_ACCESS_KEY_ID} and \env{AWS_SECRET_ACCESS_KEY} environment variables are not set.
}
\examples{
\dontrun{
# read and parse a credentials file
read_credentials()

# set environment variables from a profile
use_credentials()
}
}
\references{
\href{https://blogs.aws.amazon.com/security/post/Tx3D6U6WSFGOK2H/A-New-and-Standardized-Way-to-Manage-Credentials-in-the-AWS-SDKs}{Amazon blog post describing the format}
}
\seealso{
\code{\link{signature_v2_auth}}, \code{\link{locate_credentials}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locate_credentials.R
\name{locate_credentials}
\alias{locate_credentials}
\title{Locate AWS Credentials}
\usage{
locate_credentials(
  key = NULL,
  secret = NULL,
  session_token = NULL,
  region = NULL,
  file = Sys.getenv("AWS_SHARED_CREDENTIALS_FILE", default_credentials_file()),
  profile = NULL,
  default_region = getOption("cloudyr.aws.default_region", "us-east-1"),
  verbose = getOption("verbose", FALSE)
)
}
\arguments{
\item{key}{An AWS Access Key ID}

\item{secret}{An AWS Secret Access Key}

\item{session_token}{Optionally, an AWS Security Token Service (STS) temporary Session Token}

\item{region}{A character string containing the AWS region for the request. If missing, \dQuote{us-east-1} is assumed.}

\item{file}{A character string containing a path to a centralized \samp{.aws/credentials} file.}

\item{profile}{A character string specifying which profile to use from the file. By default, the profile named in \env{AWS_PROFILE} is used, otherwise the \dQuote{default} profile is used.}

\item{default_region}{A character string specifying a default string to use of no user-supplied value is found.}

\item{verbose}{A logical indicating whether to be verbose.}
}
\description{
Locate AWS credentials from likely sources
}
\details{
These functions locate values of AWS credentials (access key, secret access key, session token, and region) from likely sources. The order in which these are searched is as follows:
\enumerate{
  \item user-supplied values passed to the function
  \item environment variables, first checking for default credentials (\env{AWS_ACCESS_KEY_ID}, \env{AWS_SECRET_ACCESS_KEY}, \env{AWS_DEFAULT_REGION}, and \env{AWS_SESSION_TOKEN}); then for Web Identity Provider credentials (\env{AWS_ROLE_ARN} and \env{AWS_WEB_IDENTITY_TOKEN_FILE})
  \item an instance role (on the running ECS task from which this function is called) as identified by \code{\link[aws.ec2metadata]{metadata}}, if the aws.ec2metadata package is installed
  \item an IAM instance role (on the running EC2 instance from which this function is called) as identified by \code{\link[aws.ec2metadata]{metadata}}, if the aws.ec2metadata package is installed
  \item a profile in a local credentials dot file in the current working directory, using the profile specified by \env{AWS_PROFILE}
  \item the default profile in that local credentials file
  \item a profile in a global credentials dot file in a location set by \env{AWS_SHARED_CREDENTIALS_FILE} or defaulting typically to \file{~/.aws/credentials} (or another OS-specific location), using the profile specified by \env{AWS_PROFILE}
  \item the default profile in that global credentials file
}

If \env{AWS_ACCESS_KEY_ID} and \env{AWS_SECRET_ACCESS_KEY} environment variables are not present when the package is loaded, then \code{use_credentials} is invoked using the file specified in \env{AWS_SHARED_CREDENTIALS_FILE} (or another default location) and the profile specified in \env{AWS_PROFILE} (or, if missing, the \dQuote{default} profile).

To use this (and any cloudyr package) on AWS EC2 instances, users will also need to install the \href{https://cran.r-project.org/package=aws.ec2metadata}{aws.ec2metadata} package, which allows \code{locate_credentials} to know it is running in an instance and check for relevant values. If this package is not installed, instance metadata is not checked.

Because region is often handled slightly differently from credentials and is required for most requests (whereas some services allow anonymous requests without specifying credentials), the value of region is searched for in the same order as the above but lacking a value there fails safe with the following preference ranking of possible region values (regardless of location of other credentials):
\enumerate{
  \item a user-supplied value
  \item the \env{AWS_DEFAULT_REGION} environment variable
  \item (only on EC2 instances) a region declared in the instance metadata
  \item (if a credentials file is being used) the value specified therein
  \item the default value specified in \code{default_region} (i.e., \dQuote{us-east-1} - this can be overriden with the option \dQuote{cloudyr.aws.default_region})
}

As such, user-supplied values of \code{region} always trump any other value.
}
\seealso{
\code{\link{signature_v4}}, \code{\link{signature_v2_auth}}, \code{\link{use_credentials}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assume_role.R
\name{assume_role_with_web_identity}
\alias{assume_role_with_web_identity}
\title{Assume Role with AWS Web Identity}
\usage{
assume_role_with_web_identity(
  role_arn,
  token_file,
  base_url = Sys.getenv("AWS_STS_ENDPOINT", "https://sts.amazonaws.com"),
  session_name = NULL,
  duration = 3600,
  version = "2011-06-15",
  verbose = getOption("verbose", FALSE)
)
}
\arguments{
\item{role_arn}{A character string containing the AWS Role Amazon Resource Name (ARN). This specifies the permissions you have to access other AWS services.}

\item{token_file}{A character string containing a path to a Web Identity Token file.}

\item{base_url}{The AWS STS endpoint to use to retrieve your credentials from.}

\item{session_name}{A character string optionally specifying the name.}

\item{duration}{The expiry time on the retrieved credentials.}

\item{version}{The AWS STS specification version to use.}

\item{verbose}{A logical indicating whether to be verbose.}
}
\description{
Assume a role from a provided Web Identity Token and ARN using AWS Secure Token Service (STS).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/v2.R
\name{signature_v2_auth}
\alias{signature_v2_auth}
\title{Signature Version 2}
\usage{
signature_v2_auth(
  datetime = format(Sys.time(), "\%Y-\%m-\%dT\%H:\%M:\%S", tz = "UTC"),
  verb,
  service,
  path,
  query_args = list(),
  key = NULL,
  secret = NULL,
  region = NULL,
  force_credentials = FALSE,
  verbose = getOption("verbose", FALSE)
)
}
\arguments{
\item{datetime}{A character string containing a date in the form of \dQuote{YYYY-MM-DDTH:M:S}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.}

\item{verb}{A character string specify an HTTP verb/method (e.g., \dQuote{GET}).}

\item{service}{A character string containing the full hostname of an AWS service (e.g., \dQuote{iam.amazonaws.com}, etc.)}

\item{path}{A character string specify the path to the API endpoint.}

\item{query_args}{A list containing named query arguments.}

\item{key}{An AWS Access Key ID. If \code{NULL}, it is retrieved using \code{\link{locate_credentials}}.}

\item{secret}{An AWS Secret Access Key. If \code{NULL}, it is retrieved using \code{\link{locate_credentials}}.}

\item{region}{A character string containing the AWS region for the request. If missing, \dQuote{us-east-1} is assumed.}

\item{force_credentials}{A logical indicating whether to force use of user-supplied credentials. If \code{FALSE} (the default), \code{\link{locate_credentials}} is used to find credentials. If \code{TRUE}, user-supplied values are used regardless of their validity.}

\item{verbose}{A logical indicating whether to be verbose.}
}
\value{
A list.
}
\description{
Generates AWS Signature Version 2
}
\details{
This function generates an AWS Signature Version 2 for authorizing API requests. The function returns both an updated set of query string parameters, containing the required signature-related entries, as well as a \code{Signature} field containing the Signature string itself. Version 2 is mostly deprecated and in most cases users should rely on \code{\link{signature_v4_auth}} for Version 4 signatures instead.
}
\examples{
\dontrun{
# examples from:
# https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html

true_string <- paste0("GET\n",
"elasticmapreduce.amazonaws.com\n",
"/\n",
"AWSAccessKeyId=AKIAIOSFODNN7EXAMPLE",
"&Action=DescribeJobFlows",
"&SignatureMethod=HmacSHA256",
"&SignatureVersion=2",
"&Timestamp=2011-10-03T15\\%3A19\\%3A30",
"&Version=2009-03-31", collapse = "")
true_sig <- "i91nKc4PWAt0JJIdXwz9HxZCJDdiy6cf/Mj6vPxyYIs="

q1 <-
list(Action = "DescribeJobFlows",
     Version = "2009-03-31",
     AWSAccessKeyId = "AKIAIOSFODNN7EXAMPLE",
     SignatureVersion = "2",
     SignatureMethod = "HmacSHA256",
     Timestamp = "2011-10-03T15:19:30")

sig1 <-
signature_v2_auth(datetime = "2011-10-03T15:19:30",
                  service = "elasticmapreduce.amazonaws.com",
                  verb = "GET",
                  path = "/",
                  query_args = q1,
                  key = q1$AWSAccessKeyId,
                  secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
identical(true_string, sig1$CanonicalRequest)
identical(true_sig, sig1$Signature)

# leaving out some defaults
q2 <-
list(Action = "DescribeJobFlows",
     Version = "2009-03-31",
     Timestamp = "2011-10-03T15:19:30")
sig2 <-
signature_v2_auth(datetime = "2011-10-03T15:19:30",
                  service = "elasticmapreduce.amazonaws.com",
                  verb = "GET",
                  path = "/",
                  query_args = q2,
                  key = "AKIAIOSFODNN7EXAMPLE",
                  secret = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY")
identical(true_string, sig2$CanonicalRequest)
identical(true_sig, sig2$Signature)
}
}
\references{
\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html}{AWS General Reference: Signature Version 2 Signing Process}
}
\seealso{
\code{\link{signature_v4_auth}}, \code{\link{use_credentials}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/v4.R
\name{signature_v4_auth}
\alias{signature_v4_auth}
\title{Signature Version 4}
\usage{
signature_v4_auth(
  datetime = format(Sys.time(), "\%Y\%m\%dT\%H\%M\%SZ", tz = "UTC"),
  region = NULL,
  service,
  verb,
  action,
  query_args = list(),
  canonical_headers,
  request_body,
  signed_body = FALSE,
  key = NULL,
  secret = NULL,
  session_token = NULL,
  query = FALSE,
  algorithm = "AWS4-HMAC-SHA256",
  force_credentials = FALSE,
  verbose = getOption("verbose", FALSE)
)
}
\arguments{
\item{datetime}{A character string containing a datetime in the form of \dQuote{YYYYMMDDTHHMMSSZ}. If missing, it is generated automatically using \code{\link[base]{Sys.time}}.}

\item{region}{A character string containing the AWS region for the request. If missing, \dQuote{us-east-1} is assumed.}

\item{service}{A character string containing the AWS service (e.g., \dQuote{iam}, \dQuote{host}, \dQuote{ec2}).}

\item{verb}{A character string containing the HTTP verb being used in the request.}

\item{action}{A character string containing the API endpoint used in the request.}

\item{query_args}{A named list of character strings containing the query string values (if any) used in the API request, passed to \code{\link{canonical_request}}.}

\item{canonical_headers}{A named list of character strings containing the headers used in the request.}

\item{request_body}{The body of the HTTP request.}

\item{signed_body}{Should the body be signed}

\item{key}{An AWS Access Key ID. If \code{NULL}, it is retrieved using \code{\link{locate_credentials}}.}

\item{secret}{An AWS Secret Access Key. If \code{NULL}, it is retrieved using \code{\link{locate_credentials}}.}

\item{session_token}{Optionally, an AWS Security Token Service (STS) temporary Session Token. This is added automatically as a header to \code{canonical_headers}. See \code{\link{locate_credentials}}.}

\item{query}{A logical. Currently ignored.}

\item{algorithm}{A character string containing the hashing algorithm used in the request. Should only be \dQuote{SHA256}.}

\item{force_credentials}{A logical indicating whether to force use of user-supplied credentials. If \code{FALSE} (the default), \code{\link{locate_credentials}} is used to find credentials. If \code{TRUE}, user-supplied values are used regardless of their validity.}

\item{verbose}{A logical indicating whether to be verbose.}
}
\value{
A list of class \dQuote{aws_signature_v4}, containing the information needed to sign an AWS API request using either query string authentication or request header authentication. Specifically, the list contains:

    \item{Algorithm}{A character string containing the hashing algorithm used during the signing process (default is SHA256).}
    \item{Credential}{A character string containing an identifying credential \dQuote{scoped} to the region, date, and service of the request.}
    \item{Date}{A character string containing a YYYYMMDD-formatted date.}
    \item{SignedHeaders}{A character string containing a semicolon-separated listing of request headers used in the signature.}
    \item{Body}{The value passed to \code{request_body}.}
    \item{BodyHash}{A character string containing a SHA256 hash of the request body.}
    \item{Verb}{The value passed to \code{verb}.}
    \item{Query}{The value passed to \code{query_args}.}
    \item{Service}{The value passed to \code{service}.}
    \item{Action}{The value passed to \code{action}.}
    \item{CanonicalRequest}{A character string containing the canonical request.}
    \item{StringToSign}{A character string containing the string to sign for the request.}
    \item{Signature}{A character string containing a request signature hash.}
    \item{SignatureHeader}{A character string containing a complete Authorization header value.}
    \item{AccessKeyId}{A character string containing the access key id identified by \code{\link{locate_credentials}}.}
    \item{SecretAccessKey}{A character string containing the secret access key identified by \code{\link{locate_credentials}}.}
    \item{SessionToken}{A character string containing the session token identified by \code{\link{locate_credentials}}.}
    \item{Region}{A character string containing the region identified by \code{\link{locate_credentials}}.}

These values can either be used as query parameters in a REST-style API request, or as request headers. If authentication is supplied via query string parameters, the query string should include the following:

Action=\code{action}
&X-Amz-Algorithm=\code{Algorithm}
&X-Amz-Credential=\code{URLencode(Credentials)}
&X-Amz-Date=\code{Date}
&X-Amz-Expires=\code{timeout}
&X-Amz-SignedHeaders=\code{SignedHeaders}

where \code{action} is the API endpoint being called and \code{timeout} is a numeric value indicating when the request should expire.

If signing a request using header-based authentication, the \dQuote{Authorization} header in the request should be included with the request that looks as follows:

Authorization: \code{Algorithm} Credential=\code{Credential}, SignedHeaders=\code{SignedHeaders}, Signature=\code{Signature}

This is the value printed by default for all objects of class \dQuote{aws_signature_v4}.
}
\description{
AWS Signature Version 4 for use in query or header authorization
}
\details{
This function generates an AWS Signature Version 4 for authorizing API requests.
}
\references{
\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv.html?refid=faq_card}{AWS General Reference: Signature Version 4 Signing Process}

\href{https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-authenticating-requests.html}{Amazon S3 API Reference: Authenticating Requests (AWS Signature Version 4)}

\href{https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html}{Add the Signing Information to the Request}
}
\seealso{
\code{\link{signature_v2_auth}}, \code{\link{locate_credentials}}
}
\author{
Thomas J. Leeper <thosjleeper@gmail.com>
}
